from autolens.profiles import geometry_profiles
from autolens.profiles import light_profiles
from scipy.integrate import quad
from scipy import special
from itertools import count
import numpy as np
import numba
from numba import cfunc
from numba.types import intc, CPointer, float64
from scipy import LowLevelCallable
import inspect


def jit_integrand(integrand_function):
    jitted_function = numba.jit(integrand_function, nopython=True)
    no_args = len(inspect.getfullargspec(integrand_function).args)

    wrapped = None

    if no_args == 4:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3])
    elif no_args == 5:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3], xx[4])
    elif no_args == 6:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3], xx[4], xx[5])
    elif no_args == 7:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3], xx[4], xx[5], xx[6])
    elif no_args == 8:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3], xx[4], xx[5], xx[6], xx[7])
    elif no_args == 9:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3], xx[4], xx[5], xx[6], xx[7], xx[8])
    elif no_args == 10:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3], xx[4], xx[5], xx[6], xx[7], xx[8], xx[9])
    elif no_args == 11:
        def wrapped(n, xx):
            return jitted_function(xx[0], xx[1], xx[2], xx[3], xx[4], xx[5], xx[6], xx[7], xx[8], xx[9], xx[10])

    cf = cfunc(float64(intc, CPointer(float64)))

    return LowLevelCallable(cf(wrapped).ctypes)


class MassProfile(object):

    def surface_density_func(self, eta):
        raise NotImplementedError("surface_density_at_radius should be overridden")

    def surface_density_from_grid(self, grid):
        raise NotImplementedError("surface_density_from_coordinate_grid should be overridden")

    def potential_from_grid(self, grid):
        raise NotImplementedError("potential_from_coordinate_grid should be overridden")

    def deflections_from_grid(self, grid):
        raise NotImplementedError("deflections_from_coordinate_grid should be overridden")


# noinspection PyAbstractClass
class EllipticalMassProfile(geometry_profiles.EllipticalProfile, MassProfile):
    _ids = count()

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """
        Abstract class for elliptical mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The centre of the profile
        axis_ratio : float
            Ellipse's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of profile's ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalMassProfile, self).__init__(centre, axis_ratio, phi)
        self.axis_ratio = axis_ratio
        self.phi = phi
        self.component_number = next(self._ids)

    def tabulate_integral(self, grid, tabulate_bins):
        """Tabulate an integral over the surface density of deflection potential of a mass profile. This is used in \
        the GeneralizedNFW profile classes to speed up the integration procedure.
        
        Parameters
        -----------
        grid : mask.ImageGrid
            The grid of coordinates the potential / deflections are computed on.
        tabulate_bins : int
            The number of bins used to tabulate the integral over.
        """
        eta_min = 1.0e-4
        eta_max = 1.05 * np.max(self.grid_to_elliptical_radius(grid))

        minimum_log_eta = np.log10(eta_min)
        maximum_log_eta = np.log10(eta_max)
        bin_size = (maximum_log_eta - minimum_log_eta) / (tabulate_bins - 1)

        return eta_min, eta_max, minimum_log_eta, maximum_log_eta, bin_size

    @property
    def subscript(self):
        return ''

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi']

    def dimensionless_mass_within_circle(self, radius):
        """ Compute the mass profiles's total dimensionless mass within a circle of specified radius. This is \ 
        performed via integration of the surface density profiles and is centred on the mass profile.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.

        Returns
        -------
        dimensionless_mass : float
            The total dimensionless mass within the specified circle.
        """
        return quad(self.dimensionless_mass_integral, a=0.0, b=radius, args=(1.0,))[0]

    def dimensionless_mass_within_ellipse(self, major_axis):
        """ Compute the mass profiles's total dimensionless mass within an ellipse of specified radius. This is \
        performed via integration of the surface density profiles and is centred and rotationally aligned with the \
        mass profile.

        Parameters
        ----------
        major_axis : float
            The major-axis radius of the ellipse.

        Returns
        -------
        dimensionless_mass : float
            The total dimensionless mass within the specified circle.
        """
        return quad(self.dimensionless_mass_integral, a=0.0, b=major_axis, args=(self.axis_ratio,))[0]

    def dimensionless_mass_integral(self, x, axis_ratio):
        """Routine to integrate an elliptical light profiles - set axis ratio to 1 to compute the luminosity within a \
        circle"""
        r = x * axis_ratio
        return 2 * np.pi * r * self.surface_density_func(x)


class EllipticalPowerLaw(EllipticalMassProfile, MassProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, einstein_radius=1.0, slope=2.0):
        """
        Represents an elliptical power-law density distribution.

        Parameters
        ----------
        centre: (float, float)
            The (x,y) coordinates of the centre of the profile.
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive x-axis
        einstein_radius : float
            Einstein radius of power-law mass profile.
        slope : float
            power-law density slope of mass profile.
        """

        super(EllipticalPowerLaw, self).__init__(centre, axis_ratio, phi)
        super(MassProfile, self).__init__()

        self.einstein_radius = einstein_radius
        self.slope = slope
        self.core_radius = 0.0

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', r'\theta', r'\alpha']

    @property
    def einstein_radius_rescaled(self):
        """Rescale the einstein radius by slope and axis_ratio, to reduce its degeneracy with other mass-profiles
        parameters"""
        return ((3 - self.slope) / (1 + self.axis_ratio)) * self.einstein_radius ** (self.slope - 1)

    @geometry_profiles.transform_grid
    def surface_density_from_grid(self, grid):
        """ Calculate the projected surface density in dimensionless units at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the surface density is computed on.
        """

        surface_density_grid = np.zeros(grid.shape[0])

        grid_eta = self.grid_to_elliptical_radius(grid)

        for i in range(grid.shape[0]):
            surface_density_grid[i] = self.surface_density_func(grid_eta[i])

        return surface_density_grid

    @geometry_profiles.transform_grid
    def potential_from_grid(self, grid):
        """
        Calculate the potential at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """
        potential_grid = np.zeros(grid.shape[0])

        for i in range(grid.shape[0]):
            potential_grid[i] = quad(self.potential_func, a=0.0, b=1.0,
                                     args=(grid[i, 0], grid[i, 1], self.axis_ratio, self.slope, self.core_radius))[0]

        return self.einstein_radius_rescaled * self.axis_ratio * potential_grid

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        def calculate_deflection_component(grid, npow, index):
            deflection_grid = np.zeros(grid.shape[0])

            einstein_radius_rescaled = self.einstein_radius_rescaled

            for i in range(grid.shape[0]):
                deflection_grid[i] = self.axis_ratio * grid[i, index] * quad(self.deflection_func, a=0.0, b=1.0,
                                                                             args=(grid[i, 0], grid[i, 1], npow,
                                                                                   self.axis_ratio,
                                                                                   einstein_radius_rescaled, self.slope,
                                                                                   self.core_radius))[0]

            return deflection_grid

        deflection_x = calculate_deflection_component(grid, 0.0, 0)
        deflection_y = calculate_deflection_component(grid, 1.0, 1)

        return self.rotate_grid_from_profile(np.multiply(1.0, np.vstack((deflection_x, deflection_y)).T))

    def surface_density_func(self, radius):
        return self.einstein_radius_rescaled * radius ** (-(self.slope - 1))

    @staticmethod
    @jit_integrand
    def potential_func(u, x, y, axis_ratio, slope, core_radius):
        eta_u = np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))
        return (eta_u / u) * ((3.0 - slope) * eta_u) ** -1.0 * eta_u ** (3.0 - slope) / \
               ((1 - (1 - axis_ratio ** 2) * u) ** 0.5)

    @staticmethod
    @jit_integrand
    def deflection_func(u, x, y, npow, axis_ratio, einstein_radius_rescaled, slope, core_radius):
        eta_u = np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))
        return einstein_radius_rescaled * eta_u ** (-(slope - 1)) / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))


class SphericalPowerLaw(EllipticalPowerLaw):

    def __init__(self, centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0):
        """
        Represents a spherical power-law density distribution.

        Parameters
        ----------
        centre: (float, float)
            The (x,y) coordinates of the centre of the profile.
        einstein_radius : float
            Einstein radius of power-law mass profiles
        slope : float
            power-law density slope of mass profiles
        """

        super(SphericalPowerLaw, self).__init__(centre, 1.0, 0.0, einstein_radius, slope)

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\alpha']

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        eta = self.grid_to_radius(grid)
        deflection_r = 2.0 * self.einstein_radius_rescaled * \
                       np.divide(np.power(eta, (3.0 - self.slope)), np.multiply((3.0 - self.slope), eta))
        return self.grid_radius_to_cartesian(grid, deflection_r)


class EllipticalIsothermal(EllipticalPowerLaw):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=0.9, phi=0.0, einstein_radius=1.0):
        """
        Represents an elliptical isothermal density distribution, which is equivalent to the elliptical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        """

        super(EllipticalIsothermal, self).__init__(centre, axis_ratio, phi, einstein_radius, 2.0)

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', r'\theta']

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        try:
            factor = 2.0 * self.einstein_radius_rescaled * self.axis_ratio / np.sqrt(1 - self.axis_ratio ** 2)

            psi = np.sqrt(np.add(np.multiply(self.axis_ratio ** 2, np.square(grid[:, 0])), np.square(grid[:, 1])))

            deflection_x = np.arctan(np.divide(np.multiply(np.sqrt(1 - self.axis_ratio ** 2), grid[:, 0]), psi))
            deflection_y = np.arctanh(np.divide(np.multiply(np.sqrt(1 - self.axis_ratio ** 2), grid[:, 1]), psi))

            return self.rotate_grid_from_profile(np.multiply(factor, np.vstack((deflection_x, deflection_y)).T))
        except ZeroDivisionError:
            return self.grid_radius_to_cartesian(grid, np.full(grid.shape[0], 2.0 * self.einstein_radius_rescaled))


class SphericalIsothermal(EllipticalIsothermal):

    def __init__(self, centre=(0.0, 0.0), einstein_radius=1.0):
        """
        Represents a spherical isothermal density distribution, which is equivalent to the spherical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        einstein_radius : float
            Einstein radius of power-law mass profiles
        """

        super(SphericalIsothermal, self).__init__(centre, 1.0, 0.0, einstein_radius)

    @property
    def parameter_labels(self):
        return ['x', 'y', r'\theta']

    @geometry_profiles.transform_grid
    def potential_from_grid(self, grid):
        """
        Calculate the potential at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """
        eta = self.grid_to_elliptical_radius(grid)
        return 2.0 * self.einstein_radius_rescaled * eta

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """
        return self.grid_radius_to_cartesian(grid, np.full(grid.shape[0], 2.0 * self.einstein_radius_rescaled))


class EllipticalCoredPowerLaw(EllipticalPowerLaw):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, einstein_radius=1.0, slope=2.0, core_radius=0.05):
        """
        Represents a cored elliptical power-law density distribution

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        slope : float
            power-law density slope of mass profiles
        core_radius : float
            The radius of the inner core
        """
        super(EllipticalCoredPowerLaw, self).__init__(centre, axis_ratio, phi, einstein_radius, slope)

        self.core_radius = core_radius

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', r'\theta', r'\alpha', 'S']

    def surface_density_func(self, radius):
        return self.einstein_radius_rescaled * (self.core_radius ** 2 + radius ** 2) ** (-(self.slope - 1) / 2.0)

    @staticmethod
    @jit_integrand
    def potential_func(u, x, y, axis_ratio, slope, core_radius):
        eta = np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))
        return (eta / u) * ((3.0 - slope) * eta) ** -1.0 * \
               ((core_radius ** 2 + eta ** 2) ** ((3.0 - slope) / 2.0) -
                core_radius ** (3 - slope)) / ((1 - (1 - axis_ratio ** 2) * u) ** 0.5)

    @staticmethod
    @jit_integrand
    def deflection_func(u, x, y, npow, axis_ratio, einstein_radius_rescaled, slope, core_radius):
        eta_u = np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))
        return einstein_radius_rescaled * (core_radius ** 2 + eta_u ** 2) ** (-(slope - 1) / 2.0) \
               / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))


class SphericalCoredPowerLaw(EllipticalCoredPowerLaw):

    def __init__(self, centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0, core_radius=0.05):
        """
        Represents a cored spherical power-law density distribution

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        einstein_radius : float
            Einstein radius of power-law mass profiles
        slope : float
            power-law density slope of mass profiles
        core_radius : float
            The radius of the inner core
        """
        super(SphericalCoredPowerLaw, self).__init__(centre, 1.0, 0.0, einstein_radius, slope, core_radius)

    @property
    def parameter_labels(self):
        return ['x', 'y', r'\theta', r'\alpha', 'S']

    # TODO : Same problem as SphericalPowerLaw - however at coordinates / grid give the same defleciton_r value. Its the
    # TODO : self.grid_radius_to_cartesian method which is causing the problem.

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """
        eta = self.grid_to_radius(grid)
        deflection = np.multiply(2. * self.einstein_radius_rescaled, np.divide(
            np.add(np.power(np.add(self.core_radius ** 2, np.square(eta)), (3. - self.slope) / 2.),
                   -self.core_radius ** (3 - self.slope)), np.multiply((3. - self.slope), eta)))
        return self.grid_radius_to_cartesian(grid, deflection)


class EllipticalCoredIsothermal(EllipticalCoredPowerLaw):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, einstein_radius=1.0, core_radius=0.05):
        """
        Represents a cored elliptical isothermal density distribution, which is equivalent to the elliptical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        core_radius : float
            The radius of the inner core
        """

        super(EllipticalCoredIsothermal, self).__init__(centre, axis_ratio, phi, einstein_radius, 2.0,
                                                        core_radius)

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', r'\theta', 'S']


class SphericalCoredIsothermal(SphericalCoredPowerLaw):

    def __init__(self, centre=(0.0, 0.0), einstein_radius=1.0, core_radius=0.05):
        """
        Represents a cored spherical isothermal density distribution, which is equivalent to the elliptical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        einstein_radius : float
            Einstein radius of power-law mass profiles
        core_radius : float
            The radius of the inner core
        """

        super(SphericalCoredIsothermal, self).__init__(centre, einstein_radius, 2.0, core_radius)

    @property
    def parameter_labels(self):
        return ['x', 'y', r'\theta', 'S']


class EllipticalNFW(EllipticalMassProfile, MassProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, kappa_s=0.05, scale_radius=5.0):
        """
        The elliptical NFW profiles, used to fit the dark matter halo of the lens.

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        kappa_s : float
            The overall normalization of the dark matter halo
        scale_radius : float
            The radius containing half the light of this model_mapper
        """

        super(EllipticalNFW, self).__init__(centre, axis_ratio, phi)
        super(MassProfile, self).__init__()
        self.kappa_s = kappa_s
        self.scale_radius = scale_radius
        self.inner_slope = 1.0

    @property
    def subscript(self):
        return 'd'

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', r'\kappa', 'Rs']

    @staticmethod
    def coord_func(r):
        if r > 1:
            return (1.0 / np.sqrt(r ** 2 - 1)) * np.arctan(np.sqrt(r ** 2 - 1))
        elif r < 1:
            return (1.0 / np.sqrt(1 - r ** 2)) * np.arctanh(np.sqrt(1 - r ** 2))
        elif r == 1:
            return 1

    @geometry_profiles.transform_grid
    def surface_density_from_grid(self, grid):
        """ Calculate the projected surface density in dimensionless units at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the surface density is computed on.
        """

        surface_density_grid = np.zeros(grid.shape[0])

        grid_eta = self.grid_to_elliptical_radius(grid)

        for i in range(grid.shape[0]):
            surface_density_grid[i] = self.surface_density_func(grid_eta[i])

        return surface_density_grid

    @geometry_profiles.transform_grid
    def potential_from_grid(self, grid):
        """
        Calculate the potential at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        potential_grid = np.zeros(grid.shape[0])

        for i in range(grid.shape[0]):
            potential_grid[i] = quad(self.potential_func, a=0.0, b=1.0,
                                     args=(grid[i, 0], grid[i, 1], self.axis_ratio, self.kappa_s, self.scale_radius),
                                     epsrel=1.49e-5)[0]

        return potential_grid

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        def calculate_deflection_component(grid, npow, index):
            deflection_grid = np.zeros(grid.shape[0])

            for i in range(grid.shape[0]):
                deflection_grid[i] = self.axis_ratio * grid[i, index] * quad(self.deflection_func, a=0.0, b=1.0,
                                                                             args=(grid[i, 0], grid[i, 1], npow,
                                                                                   self.axis_ratio, self.kappa_s,
                                                                                   self.scale_radius))[0]

            return deflection_grid

        deflection_x = calculate_deflection_component(grid, 0.0, 0)
        deflection_y = calculate_deflection_component(grid, 1.0, 1)

        return self.rotate_grid_from_profile(np.multiply(1.0, np.vstack((deflection_x, deflection_y)).T))

    def surface_density_func(self, radius):
        radius = (1.0 / self.scale_radius) * radius
        return 2.0 * self.kappa_s * (1 - self.coord_func(radius)) / (radius ** 2 - 1)

    @staticmethod
    @jit_integrand
    def potential_func(u, x, y, axis_ratio, kappa_s, scale_radius):

        eta_u = (1.0 / scale_radius) * np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))

        if eta_u > 1:
            eta_u_2 = (1.0 / np.sqrt(eta_u ** 2 - 1)) * np.arctan(np.sqrt(eta_u ** 2 - 1))
        elif eta_u < 1:
            eta_u_2 = (1.0 / np.sqrt(1 - eta_u ** 2)) * np.arctanh(np.sqrt(1 - eta_u ** 2))
        else:
            eta_u_2 = 1

        return 4.0 * kappa_s * scale_radius * (axis_ratio / 2.0) * (eta_u / u) * (
                (np.log(eta_u / 2.0) + eta_u_2) / eta_u) / (
                       (1 - (1 - axis_ratio ** 2) * u) ** 0.5)

    @staticmethod
    @jit_integrand
    def deflection_func(u, x, y, npow, axis_ratio, kappa_s, scale_radius):

        eta_u = (1.0 / scale_radius) * np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))

        if eta_u > 1:
            eta_u_2 = (1.0 / np.sqrt(eta_u ** 2 - 1)) * np.arctan(np.sqrt(eta_u ** 2 - 1))
        elif eta_u < 1:
            eta_u_2 = (1.0 / np.sqrt(1 - eta_u ** 2)) * np.arctanh(np.sqrt(1 - eta_u ** 2))
        else:
            eta_u_2 = 1

        return 2.0 * kappa_s * (1 - eta_u_2) / (eta_u ** 2 - 1) / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))


class SphericalNFW(EllipticalNFW):

    def __init__(self, centre=(0.0, 0.0), kappa_s=0.05, scale_radius=5.0):
        """
        The spherical NFW profiles, used to fit the dark matter halo of the lens.

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        kappa_s : float
            The overall normalization of the dark matter halo
        scale_radius : float
            The radius containing half the light of this model_mapper
        """

        super(SphericalNFW, self).__init__(centre, 1.0, 0.0, kappa_s, scale_radius)

    @property
    def parameter_labels(self):
        return ['x', 'y', r'\kappa', 'Rs']

    # TODO : The 'func' routines require a different input to the elliptical cases, meaning they cannot be overridden.
    # TODO : Should be able to refactor code to deal with this nicely, but will wait until we're clear on numba.

    # TODO : Make this use numpy arthimitic

    @geometry_profiles.transform_grid
    def potential_from_grid(self, grid):
        """
        Calculate the potential at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """
        eta = (1.0 / self.scale_radius) * self.grid_to_radius(grid)
        return 2.0 * self.scale_radius * self.kappa_s * self.potential_func_sph(eta)

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """
        eta = np.multiply(1. / self.scale_radius, self.grid_to_radius(grid))
        deflection_r = np.multiply(4. * self.kappa_s * self.scale_radius, self.deflection_func_sph(eta))

        return self.grid_radius_to_cartesian(grid, deflection_r)

    @staticmethod
    def potential_func_sph(eta):
        return ((np.log(eta / 2.0)) ** 2) - (np.arctanh(np.sqrt(1 - eta ** 2))) ** 2

    @staticmethod
    def deflection_func_sph(eta):
        conditional_eta = np.copy(eta)
        conditional_eta[eta > 1] = np.multiply(np.divide(1.0, np.sqrt(np.add(np.square(eta[eta > 1]), - 1))),
                                               np.arctan(np.sqrt(np.add(np.square(eta[eta > 1]), - 1))))
        conditional_eta[eta < 1] = np.multiply(np.divide(1.0, np.sqrt(np.add(1, - np.square(eta[eta < 1])))),
                                               np.arctanh(np.sqrt(np.add(1, - np.square(eta[eta < 1])))))

        return np.divide(np.add(np.log(np.divide(eta, 2.)), conditional_eta), eta)


class EllipticalGeneralizedNFW(EllipticalNFW):
    epsrel = 1.49e-5

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, kappa_s=0.05, inner_slope=1.0, scale_radius=5.0):
        """
        The elliptical NFW profiles, used to fit the dark matter halo of the lens.

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        kappa_s : float
            The overall normalization of the dark matter halo
        inner_slope : float
            The inner slope of the dark matter halo
        scale_radius : float
            The radius containing half the light of this model_mapper
        """

        super(EllipticalGeneralizedNFW, self).__init__(centre, axis_ratio, phi, kappa_s, scale_radius)
        self.inner_slope = inner_slope

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', r'\kappa', r'\gamma' 'Rs']

    @geometry_profiles.transform_grid
    def potential_from_grid(self, grid, tabulate_bins=1000):
        """
        Calculate the potential at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        @jit_integrand
        def deflection_integrand(x, kappa_radius, scale_radius, inner_slope):
            return (x + kappa_radius / scale_radius) ** (inner_slope - 3) * ((1 - np.sqrt(1 - x ** 2)) / x)

        eta_min, eta_max, minimum_log_eta, maximum_log_eta, bin_size = self.tabulate_integral(grid, tabulate_bins)

        potential_grid = np.zeros(grid.shape[0])

        deflection_integral = np.zeros((tabulate_bins))

        for i in range(tabulate_bins):
            eta = 10. ** (minimum_log_eta + (i - 1) * bin_size)

            integral = \
                quad(deflection_integrand, a=0.0, b=1.0, args=(eta, self.scale_radius, self.inner_slope),
                     epsrel=EllipticalGeneralizedNFW.epsrel)[0]

            deflection_integral[i] = ((eta / self.scale_radius) ** (2 - self.inner_slope)) * (
                    (1.0 / (3 - self.inner_slope)) *
                    special.hyp2f1(3 - self.inner_slope, 3 - self.inner_slope, 4 - self.inner_slope,
                                   - (eta / self.scale_radius)) + integral)

        for i in range(grid.shape[0]):
            potential_grid[i] = (2.0 * self.kappa_s * self.axis_ratio) * \
                                quad(self.potential_func, a=0.0, b=1.0, args=(grid[i, 0], grid[i, 1],
                                                                              self.axis_ratio, minimum_log_eta,
                                                                              maximum_log_eta, tabulate_bins,
                                                                              deflection_integral),
                                     epsrel=EllipticalGeneralizedNFW.epsrel)[0]

        return potential_grid

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid, tabulate_bins=1000):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        @jit_integrand
        def surface_density_integrand(x, kappa_radius, scale_radius, inner_slope):
            return (3 - inner_slope) * (x + kappa_radius / scale_radius) ** (inner_slope - 4) * (1 - np.sqrt(1 - x * x))

        def calculate_deflection_component(grid, npow, index):

            deflection_grid = np.zeros(grid.shape[0])

            for i in range(grid.shape[0]):
                deflection_grid[i] = 2.0 * self.kappa_s * self.axis_ratio * grid[i, index] * quad(self.deflection_func,
                                                                                                  a=0.0, b=1.0, args=(
                        grid[i, 0], grid[i, 1], npow, self.axis_ratio, minimum_log_eta, maximum_log_eta,
                        tabulate_bins, surface_density_integral), epsrel=EllipticalGeneralizedNFW.epsrel)[0]

            return deflection_grid

        eta_min, eta_max, minimum_log_eta, maximum_log_eta, bin_size = self.tabulate_integral(grid, tabulate_bins)

        surface_density_integral = np.zeros((tabulate_bins))

        for i in range(tabulate_bins):
            eta = 10. ** (minimum_log_eta + (i - 1) * bin_size)

            integral = quad(surface_density_integrand, a=0.0, b=1.0, args=(eta, self.scale_radius,
                                                                           self.inner_slope),
                            epsrel=EllipticalGeneralizedNFW.epsrel)[0]

            surface_density_integral[i] = ((eta / self.scale_radius) ** (1 - self.inner_slope)) * \
                                          (((1 + eta / self.scale_radius) ** (self.inner_slope - 3)) + integral)

        deflection_x = calculate_deflection_component(grid, 0.0, 0)
        deflection_y = calculate_deflection_component(grid, 1.0, 1)

        return self.rotate_grid_from_profile(np.multiply(1.0, np.vstack((deflection_x, deflection_y)).T))

    def surface_density_func(self, radius):

        def integral_y(y, eta):
            return (y + eta) ** (self.inner_slope - 4) * (1 - np.sqrt(1 - y ** 2))

        radius = (1.0 / self.scale_radius) * radius
        integral_y = quad(integral_y, a=0.0, b=1.0, args=radius, epsrel=EllipticalGeneralizedNFW.epsrel)[0]

        return 2.0 * self.kappa_s * (radius ** (1 - self.inner_slope)) * (
                (1 + radius) ** (self.inner_slope - 3) + ((3 - self.inner_slope) * integral_y))

    @staticmethod
    # TODO : Decorator needs to know that potential_integral is 1D array
    #    @jit_integrand
    def potential_func(u, x, y, axis_ratio, minimum_log_eta, maximum_log_eta, tabulate_bins, potential_integral):
        eta_u = np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))
        bin_size = (maximum_log_eta - minimum_log_eta) / (tabulate_bins - 1)
        i = 1 + int((np.log10(eta_u) - minimum_log_eta) / bin_size)
        r1 = 10. ** (minimum_log_eta + (i - 1) * bin_size)
        r2 = r1 * 10. ** bin_size
        phi = potential_integral[i] + (potential_integral[i + 1] - potential_integral[i]) \
              * (eta_u - r1) / (r2 - r1)
        return eta_u * (phi / u) / (1.0 - (1.0 - axis_ratio ** 2) * u) ** (0.5)

    @staticmethod
    # TODO : Decorator needs to know that surface_density_integral is 1D array
    #    @jit_integrand
    def deflection_func(u, x, y, npow, axis_ratio, minimum_log_eta, maximum_log_eta, tabulate_bins,
                        surface_density_integral):

        eta_u = np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))
        bin_size = (maximum_log_eta - minimum_log_eta) / (tabulate_bins - 1)
        i = 1 + int((np.log10(eta_u) - minimum_log_eta) / bin_size)
        r1 = 10. ** (minimum_log_eta + (i - 1) * bin_size)
        r2 = r1 * 10. ** bin_size
        kap = surface_density_integral[i] + (surface_density_integral[i + 1] - surface_density_integral[i]) \
              * (eta_u - r1) / (r2 - r1)
        return kap / (1.0 - (1.0 - axis_ratio ** 2) * u) ** (npow + 0.5)


class SphericalGeneralizedNFW(EllipticalGeneralizedNFW):

    def __init__(self, centre=(0.0, 0.0), kappa_s=0.05, inner_slope=1.0, scale_radius=5.0):
        """
        The spherical NFW profiles, used to fit the dark matter halo of the lens.

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        kappa_s : float
            The overall normalization of the dark matter halo
        inner_slope : float
            The inner slope of the dark matter halo
        scale_radius : float
            The radius containing half the light of this model_mapper
        """

        super(SphericalGeneralizedNFW, self).__init__(centre, 1.0, 0.0, kappa_s, inner_slope, scale_radius)

    @property
    def parameter_labels(self):
        return ['x', 'y', r'\kappa', r'\gamma' 'Rs']

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid, **kwargs):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        eta = np.multiply(1. / self.scale_radius, self.grid_to_radius(grid))

        deflection_grid = np.zeros(grid.shape[0])

        for i in range(grid.shape[0]):
            deflection_grid[i] = np.multiply(4. * self.kappa_s * self.scale_radius, self.deflection_func_sph(eta[i]))

        return self.grid_radius_to_cartesian(grid, deflection_grid)

    @staticmethod
    def deflection_integrand(y, eta, inner_slope):
        return (y + eta) ** (inner_slope - 3) * ((1 - np.sqrt(1 - y ** 2)) / y)

    def deflection_func_sph(self, eta):
        integral_y_2 = quad(self.deflection_integrand, a=0.0, b=1.0, args=(eta, self.inner_slope), epsrel=1.49e-6)[0]
        return eta ** (2 - self.inner_slope) * ((1.0 / (3 - self.inner_slope)) *
                                                special.hyp2f1(3 - self.inner_slope, 3 - self.inner_slope,
                                                               4 - self.inner_slope, -eta) + integral_y_2)

    #     deflection_r = np.multiply(4. * self.kappa_s * self.scale_radius, self.deflection_func_sph_grid(grid))
    #
    # def deflection_func_sph_grid(self, grid):
    #     # TODO
    #     pass


class EllipticalSersicMass(light_profiles.EllipticalSersic, EllipticalMassProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0, mass_to_light_ratio=1.0):
        """
        The Sersic mass profile, the mass profiles of the light profiles that are used to fit and subtract the lens \
        galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The radius containing half the light of this model_mapper
        sersic_index : Int
            The concentration of the light profiles
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        """
        super(EllipticalSersicMass, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, sersic_index)
        super(EllipticalMassProfile, self).__init__(centre, axis_ratio, phi)
        self.mass_to_light_ratio = mass_to_light_ratio

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R', 'n', r'\Psi']

    @geometry_profiles.transform_grid
    def surface_density_from_grid(self, grid):
        """ Calculate the projected surface density in dimensionless units at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the surface density is computed on.
        """
        return self.surface_density_func(self.grid_to_eccentric_radii(grid))

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        def calculate_deflection_component(grid, npow, index):
            deflection_grid = np.zeros(grid.shape[0])

            sersic_constant = self.sersic_constant

            for i in range(grid.shape[0]):
                deflection_grid[i] = self.axis_ratio * grid[i, index] * quad(self.deflection_func, a=0.0, b=1.0,
                                                                             args=(grid[i, 0], grid[i, 1], npow,
                                                                                   self.axis_ratio, self.intensity,
                                                                                   self.sersic_index,
                                                                                   self.effective_radius,
                                                                                   self.mass_to_light_ratio,
                                                                                   sersic_constant))[0]

            return deflection_grid

        deflection_x = calculate_deflection_component(grid, 0.0, 0)
        deflection_y = calculate_deflection_component(grid, 1.0, 1)

        return self.rotate_grid_from_profile(np.multiply(1.0, np.vstack((deflection_x, deflection_y)).T))

    def surface_density_func(self, radius):
        return self.mass_to_light_ratio * self.intensity_at_radius(radius)

    @staticmethod
    @jit_integrand
    def deflection_func(u, x, y, npow, axis_ratio, intensity, sersic_index, effective_radius, mass_to_light_ratio,
                        sersic_constant):
        eta_u = np.sqrt(axis_ratio) * np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))

        return mass_to_light_ratio * intensity * \
               np.exp(-sersic_constant * (((eta_u / effective_radius) ** (1. / sersic_index)) - 1)) \
               / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))


class EllipticalExponentialMass(EllipticalSersicMass):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 mass_to_light_ratio=1.0):
        """
        The EllipticalExponentialMass mass profile, the mass profiles of the light profiles that are used to fit and
        subtract the lens galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The radius containing half the light of this model_mapper
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        """
        super(EllipticalExponentialMass, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, 1.0,
                                                        mass_to_light_ratio)

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R', r'\Psi']

    @classmethod
    def from_exponential_light_profile(cls, exponential_light_profile, mass_to_light_ratio):
        return EllipticalExponentialMass.from_profile(exponential_light_profile,
                                                      mass_to_light_ratio=mass_to_light_ratio)


class EllipticalDevVaucouleursMass(EllipticalSersicMass):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 mass_to_light_ratio=1.0):
        """
        The EllipticalDevVaucouleursMass mass profile, the mass profiles of the light profiles that are used to fit and
        subtract the lens galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The radius containing half the light of this model_mapper
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        """
        super(EllipticalDevVaucouleursMass, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, 4.0,
                                                           mass_to_light_ratio)

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R', r'\Psi']

    @classmethod
    def from_dev_vaucouleurs_light_profile(cls, dev_vaucouleurs_light_profile, mass_to_light_ratio):
        return EllipticalDevVaucouleursMass.from_profile(dev_vaucouleurs_light_profile,
                                                         mass_to_light_ratio=mass_to_light_ratio)


class EllipticalSersicMassRadialGradient(EllipticalSersicMass):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0, mass_to_light_ratio=1.0, mass_to_light_gradient=0.0):
        """
        Setup a Sersic mass and light profiles.

        Parameters
        ----------
        centre: (float, float)
            The centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The radius containing half the light of this model_mapper
        sersic_index : Int
            The concentration of the light profiles
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        mass_to_light_gradient : float
            The mass-to-light radial gradient.
        """
        super(EllipticalSersicMassRadialGradient, self).__init__(centre, axis_ratio, phi, intensity, effective_radius,
                                                                 sersic_index, mass_to_light_ratio)
        self.mass_to_light_gradient = mass_to_light_gradient

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R', 'n', r'\Psi', r'\Tau']

    @geometry_profiles.transform_grid
    def surface_density_from_grid(self, grid):
        """ Calculate the projected surface density in dimensionless units at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the surface density is computed on.
        """
        return self.surface_density_func(self.grid_to_eccentric_radii(grid))

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        def calculate_deflection_component(grid, npow, index):
            deflection_grid = np.zeros(grid.shape[0])

            sersic_constant = self.sersic_constant

            for i in range(grid.shape[0]):
                deflection_grid[i] = self.axis_ratio * grid[i, index] * quad(self.deflection_func, a=0.0, b=1.0,
                                                                             args=(grid[i, 0], grid[i, 1], npow,
                                                                                   self.axis_ratio, self.intensity,
                                                                                   self.sersic_index,
                                                                                   self.effective_radius,
                                                                                   self.mass_to_light_ratio,
                                                                                   self.mass_to_light_gradient,
                                                                                   sersic_constant))[0]

            return deflection_grid

        deflection_x = calculate_deflection_component(grid, 0.0, 0)
        deflection_y = calculate_deflection_component(grid, 1.0, 1)

        return self.rotate_grid_from_profile(np.multiply(1.0, np.vstack((deflection_x, deflection_y)).T))

    def surface_density_func(self, radius):
        return self.mass_to_light_ratio * (
                ((self.axis_ratio * radius) / self.effective_radius) ** -self.mass_to_light_gradient) \
               * self.intensity_at_radius(radius)

    @staticmethod
    @jit_integrand
    def deflection_func(u, x, y, npow, axis_ratio, intensity, sersic_index, effective_radius, mass_to_light_ratio,
                        mass_to_light_gradient, sersic_constant):
        eta_u = np.sqrt(axis_ratio) * np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))

        return mass_to_light_ratio * (((axis_ratio * eta_u) / effective_radius) ** -mass_to_light_gradient) * \
               intensity * np.exp(-sersic_constant * (((eta_u / effective_radius) ** (1. / sersic_index)) - 1)) \
               / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))


class ExternalShear(geometry_profiles.EllipticalProfile, MassProfile):
    def __init__(self, magnitude=0.2, phi=0.0):
        """
        An external shear term, to model the line-of-sight contribution of other galaxies / satellites.

        Parameters
        ----------
        magnitude : float
            The overall magnitude of the shear (gamma).
        phi : float
            The rotation axis of the shear.
        """

        super(ExternalShear, self).__init__(centre=(0.0, 0.0), phi=phi, axis_ratio=1.0)
        self.magnitude = magnitude

    @property
    def subscript(self):
        return 'sh'

    @property
    def parameter_labels(self):
        return [r'\gamma', r'\theta']

    @geometry_profiles.transform_grid
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """
        deflection_x = np.multiply(self.magnitude, grid[:, 0])
        deflection_y = -np.multiply(self.magnitude, grid[:, 1])
        return self.rotate_grid_from_profile(np.vstack((deflection_x, deflection_y)).T)



from autolens.analysis import galaxy
from autolens.profiles import mass_profiles, light_profiles

import pytest
import numpy as np


class TestLightProfiles(object):

    class TestIntensity:

        def test__one_profile_galaxy__intensity_is_same_individual_profile(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                     effective_radius=0.6, sersic_index=4.0)

            sersic_intensity = sersic.intensity_from_grid(grid=np.array([[1.05, -0.55]]))

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=1.0,
                                                                                          effective_radius=0.6,
                                                                                          sersic_index=4.0))

            galaxy_sersic_intensity = galaxy_sersic.intensity_from_grid(np.array([[1.05, -0.55]]))

            assert sersic_intensity == galaxy_sersic_intensity

        def test__two_profile_galaxy__intensity_is_sum_of_individual_profiles(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            intensity = sersic_1.intensity_from_grid(np.array([[1.05, -0.55]]))
            intensity += sersic_2.intensity_from_grid(np.array([[1.05, -0.55]]))

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity = galaxy_sersic.intensity_from_grid(np.array([[1.05, -0.55]]))

            assert intensity == galaxy_intensity

        def test__three_profile_galaxy__intensity_is_sum_of_individual_profiles(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            sersic_3 = light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0, intensity=2.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            intensity = sersic_1.intensity_from_grid(np.array([[1.05, -0.55]]))
            intensity += sersic_2.intensity_from_grid(np.array([[1.05, -0.55]]))
            intensity += sersic_3.intensity_from_grid(np.array([[1.05, -0.55]]))

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0),
                                          light_profile_3=light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0,
                                                                                          intensity=2.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity = galaxy_sersic.intensity_from_grid(np.array([[1.05, -0.55]]))

            assert intensity == galaxy_intensity

        def test__three_profile_galaxy__individual_intensities_can_be_extracted(self):

            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            sersic_3 = light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0, intensity=2.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            intensity_1 = sersic_1.intensity_from_grid(np.array([[1.05, -0.55]]))
            intensity_2 = sersic_2.intensity_from_grid(np.array([[1.05, -0.55]]))
            intensity_3 = sersic_3.intensity_from_grid(np.array([[1.05, -0.55]]))

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0),
                                          light_profile_3=light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0,
                                                                                          intensity=2.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity = galaxy_sersic.intensity_from_grid_individual(np.array([[1.05, -0.55]]))

            assert intensity_1 == galaxy_intensity[0]
            assert intensity_2 == galaxy_intensity[1]
            assert intensity_3 == galaxy_intensity[2]

    class TestLuminosityWithinCircle:

        def test__one_profile_galaxy__integral_is_same_as_individual_profile(self):
            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=1.0)

            integral_radius = 5.5

            intensity_integral = sersic.luminosity_within_circle(radius=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_circle(radius=integral_radius)

            assert intensity_integral == galaxy_intensity_integral

        def test__two_profile_galaxy__integral_is_sum_of_individual_profiles(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            integral_radius = 5.5

            intensity_integral = sersic_1.luminosity_within_circle(radius=integral_radius)
            intensity_integral += sersic_2.luminosity_within_circle(radius=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_circle(radius=integral_radius)

            assert intensity_integral == galaxy_intensity_integral

        def test__three_profile_galaxy__integral_is_sum_of_individual_profiles(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            sersic_3 = light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0, intensity=2.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            integral_radius = 5.5

            intensity_integral = sersic_1.luminosity_within_circle(radius=integral_radius)
            intensity_integral += sersic_2.luminosity_within_circle(radius=integral_radius)
            intensity_integral += sersic_3.luminosity_within_circle(radius=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0),
                                          light_profile_3=light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0,
                                                                                          intensity=2.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_circle(radius=integral_radius)

            assert intensity_integral == galaxy_intensity_integral

        def test__three_profile_galaxy__individual_integrals_can_be_extracted(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            sersic_3 = light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0, intensity=2.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            integral_radius = 5.5

            intensity_integral_1 = sersic_1.luminosity_within_circle(radius=integral_radius)
            intensity_integral_2 = sersic_2.luminosity_within_circle(radius=integral_radius)
            intensity_integral_3 = sersic_3.luminosity_within_circle(radius=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0),
                                          light_profile_3=light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0,
                                                                                          intensity=2.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_circle_individual(
                radius=integral_radius)

            assert intensity_integral_1 == galaxy_intensity_integral[0]
            assert intensity_integral_2 == galaxy_intensity_integral[1]
            assert intensity_integral_3 == galaxy_intensity_integral[2]

    class TestLuminosityWithinEllipse:

        def test__one_profile_galaxy__integral_is_same_as_individual_profile(self):
            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=1.0)

            integral_radius = 0.5

            intensity_integral = sersic.luminosity_within_ellipse(major_axis=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=0.5,
                                                                                          phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_ellipse(major_axis=integral_radius)

            assert intensity_integral == galaxy_intensity_integral

        def test__two_profile_galaxy__integral_is_sum_of_individual_profiles(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            integral_radius = 5.5

            intensity_integral = sersic_1.luminosity_within_ellipse(major_axis=integral_radius)
            intensity_integral += sersic_2.luminosity_within_ellipse(major_axis=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_ellipse(major_axis=integral_radius)

            assert intensity_integral == galaxy_intensity_integral

        def test__three_profile_galaxy__integral_is_sum_of_individual_profiles(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            sersic_3 = light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0, intensity=2.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            integral_radius = 5.5

            intensity_integral = sersic_1.luminosity_within_ellipse(major_axis=integral_radius)
            intensity_integral += sersic_2.luminosity_within_ellipse(major_axis=integral_radius)
            intensity_integral += sersic_3.luminosity_within_ellipse(major_axis=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0),
                                          light_profile_3=light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0,
                                                                                          intensity=2.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_ellipse(major_axis=integral_radius)

            assert intensity_integral == galaxy_intensity_integral

        def test__three_profile_galaxy__individual_integrals_can_be_extracted(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                       effective_radius=2.0,
                                                       sersic_index=1.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=7.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            sersic_3 = light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0, intensity=2.0,
                                                       effective_radius=3.0,
                                                       sersic_index=2.0)

            integral_radius = 5.5

            intensity_integral_1 = sersic_1.luminosity_within_ellipse(major_axis=integral_radius)
            intensity_integral_2 = sersic_2.luminosity_within_ellipse(major_axis=integral_radius)
            intensity_integral_3 = sersic_3.luminosity_within_ellipse(major_axis=integral_radius)

            galaxy_sersic = galaxy.Galaxy(redshift=0.5,
                                          light_profile_1=light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0,
                                                                                          intensity=3.0,
                                                                                          effective_radius=2.0,
                                                                                          sersic_index=1.0),
                                          light_profile_2=light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0,
                                                                                          intensity=7.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0),
                                          light_profile_3=light_profiles.EllipticalSersic(axis_ratio=0.8, phi=50.0,
                                                                                          intensity=2.0,
                                                                                          effective_radius=3.0,
                                                                                          sersic_index=2.0))

            galaxy_intensity_integral = galaxy_sersic.luminosity_within_ellipse_individual(
                major_axis=integral_radius)

            assert intensity_integral_1 == galaxy_intensity_integral[0]
            assert intensity_integral_2 == galaxy_intensity_integral[1]
            assert intensity_integral_3 == galaxy_intensity_integral[2]

    class TestSymmetricProfiles(object):

        def test_1d_symmetry(self):

            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                       sersic_index=4.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                       sersic_index=4.0, centre=(100, 0))

            galaxy_sersic = galaxy.Galaxy(redshift=0.5, light_profile_1=sersic_1, light_profile_2=sersic_2)

            assert galaxy_sersic.intensity_from_grid(
                np.array([[0.0, 0.0]])) == galaxy_sersic.intensity_from_grid(np.array([[100.0, 0.0]]))
            assert galaxy_sersic.intensity_from_grid(
                np.array([[49.0, 0.0]])) == galaxy_sersic.intensity_from_grid(np.array([[51.0, 0.0]]))

        def test_2d_symmetry(self):
            sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                       sersic_index=4.0)

            sersic_2 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                       sersic_index=4.0, centre=(100, 0))

            sersic_3 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                       sersic_index=4.0, centre=(0, 100))

            sersic_4 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                       sersic_index=4.0, centre=(100, 100))

            galaxy_sersic = galaxy.Galaxy(redshift=0.5, light_profile_1=sersic_1, light_profile_2=sersic_2,
                                          light_profile_3=sersic_3, light_profile_4=sersic_4)

            assert galaxy_sersic.intensity_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                galaxy_sersic.intensity_from_grid(np.array([[51.0, 0.0]])), 1e-5)

            assert galaxy_sersic.intensity_from_grid(np.array([[0.0, 49.0]])) == pytest.approx(
                galaxy_sersic.intensity_from_grid(np.array([[0.0, 51.0]])), 1e-5)

            assert galaxy_sersic.intensity_from_grid(np.array([[100.0, 49.0]])) == pytest.approx(
                galaxy_sersic.intensity_from_grid(np.array([[100.0, 51.0]])), 1e-5)

            assert galaxy_sersic.intensity_from_grid(np.array([[49.0, 49.0]])) == pytest.approx(
                galaxy_sersic.intensity_from_grid(np.array([[51.0, 51.0]])), 1e-5)


@pytest.fixture(name="sie_1")
def make_sie_1():
    return mass_profiles.EllipticalIsothermal(axis_ratio=0.8, phi=10.0, einstein_radius=1.0)


@pytest.fixture(name="sie_2")
def make_sie_2():
    return mass_profiles.EllipticalIsothermal(axis_ratio=0.6, phi=30.0, einstein_radius=1.2)


@pytest.fixture(name="sie_3")
def make_sie_3():
    return mass_profiles.EllipticalIsothermal(axis_ratio=0.9, phi=130.0, einstein_radius=1.6)


@pytest.fixture(name="galaxy_sie_1")
def make_galaxy_sie_1(sie_1):
    return galaxy.Galaxy(redshift=0.5, mass_profile_1=sie_1)


@pytest.fixture(name="galaxy_sie_2")
def make_galaxy_sie_2(sie_1, sie_2):
    return galaxy.Galaxy(redshift=0.5, mass_profile_1=sie_1, mass_profile_2=sie_2)


@pytest.fixture(name="galaxy_sie_3")
def make_galaxy_sie_3(sie_1, sie_2, sie_3):
    return galaxy.Galaxy(redshift=0.5, mass_profile_1=sie_1, mass_profile_2=sie_2, mass_profile_3=sie_3)


class TestMassProfiles(object):

    class TestSurfaceDensity:

        def test__one_profile_galaxy__surface_density_is_same_individual_profile(self, sie_1, galaxy_sie_1):

            sie_surface_density = sie_1.surface_density_from_grid(np.array([[1.05, -0.55]]))

            galaxy_sie_surface_density = galaxy_sie_1.surface_density_from_grid(np.array([[1.05, -0.55]]))

            assert sie_surface_density == galaxy_sie_surface_density

        def test__two_profile_galaxy__surface_density_is_sum_of_individual_profiles(self, sie_1, sie_2, galaxy_sie_2):
            surface_density = sie_1.surface_density_from_grid(np.array([[1.05, -0.55]]))
            surface_density += sie_2.surface_density_from_grid(np.array([[1.05, -0.55]]))

            galaxy_surface_density = galaxy_sie_2.surface_density_from_grid(np.array([[1.05, -0.55]]))

            assert surface_density == galaxy_surface_density

        def test__three_profile_galaxy__surface_density_is_sum_of_individual_profiles(self, sie_1, sie_2, sie_3,
                                                                                      galaxy_sie_3):
            surface_density = sie_1.surface_density_from_grid(np.array([[1.05, -0.55]]))
            surface_density += sie_2.surface_density_from_grid(np.array([[1.05, -0.55]]))
            surface_density += sie_3.surface_density_from_grid(np.array([[1.05, -0.55]]))

            galaxy_surface_density = galaxy_sie_3.surface_density_from_grid(np.array([[1.05, -0.55]]))

            assert surface_density == galaxy_surface_density

        def test__three_profile_galaxy__individual_surface_densities_can_be_extracted(self, sie_1, sie_2, sie_3,
                                                                                      galaxy_sie_3):

            surface_density_1 = sie_1.surface_density_from_grid(np.array([[1.05, -0.55]]))
            surface_density_2 = sie_2.surface_density_from_grid(np.array([[1.05, -0.55]]))
            surface_density_3 = sie_3.surface_density_from_grid(np.array([[1.05, -0.55]]))

            galaxy_surface_density = galaxy_sie_3.surface_density_from_grid_individual(np.array([[1.05, -0.55]]))

            assert surface_density_1 == galaxy_surface_density[0]
            assert surface_density_2 == galaxy_surface_density[1]
            assert surface_density_3 == galaxy_surface_density[2]

    class TestPotential:

        def test__one_profile_galaxy__potential_is_same_individual_profile(self, sie_1, galaxy_sie_1):

            sie_potential = sie_1.potential_from_grid(np.array([[1.05, -0.55]]))

            galaxy_sie_potential = galaxy_sie_1.potential_from_grid(np.array([[1.05, -0.55]]))

            assert sie_potential == galaxy_sie_potential

        def test__two_profile_galaxy__potential_is_sum_of_individual_profiles(self, sie_1, sie_2, galaxy_sie_2):
            potential = sie_1.potential_from_grid(np.array([[1.05, -0.55]]))
            potential += sie_2.potential_from_grid(np.array([[1.05, -0.55]]))

            galaxy_potential = galaxy_sie_2.potential_from_grid(np.array([[1.05, -0.55]]))

            assert potential == galaxy_potential

        def test__three_profile_galaxy__potential_is_sum_of_individual_profiles(self, sie_1, sie_2, sie_3,
                                                                                galaxy_sie_3):
            potential = sie_1.potential_from_grid(np.array([[1.05, -0.55]]))
            potential += sie_2.potential_from_grid(np.array([[1.05, -0.55]]))
            potential += sie_3.potential_from_grid(np.array([[1.05, -0.55]]))

            galaxy_potential = galaxy_sie_3.potential_from_grid(np.array([[1.05, -0.55]]))

            assert potential == galaxy_potential

        def test__three_profile_galaxy__individual_potentials_can_be_extracted(self, sie_1, sie_2, sie_3,
                                                                               galaxy_sie_3):

            potential_1 = sie_1.potential_from_grid(np.array([[1.05, -0.55]]))
            potential_2 = sie_2.potential_from_grid(np.array([[1.05, -0.55]]))
            potential_3 = sie_3.potential_from_grid(np.array([[1.05, -0.55]]))

            galaxy_potential = galaxy_sie_3.potential_from_grid_individual(np.array([[1.05, -0.55]]))

            assert potential_1 == galaxy_potential[0]
            assert potential_2 == galaxy_potential[1]
            assert potential_3 == galaxy_potential[2]

    class TestDeflectionAngles:

        def test__one_profile_galaxy__deflection_angles_is_same_individual_profile(self, sie_1, galaxy_sie_1):

            sie_deflection_angles = sie_1.deflections_from_grid(np.array([[1.05, -0.55]]))

            galaxy_sie_deflection_angles = galaxy_sie_1.deflections_from_grid(np.array([[1.05, -0.55]]))

            assert sie_deflection_angles[0,0] == galaxy_sie_deflection_angles[0,0]
            assert sie_deflection_angles[0,1] == galaxy_sie_deflection_angles[0,1]

        def test__two_profile_galaxy__deflection_angles_is_sum_of_individual_profiles(self):

            sie_1 = mass_profiles.EllipticalIsothermal(axis_ratio=0.8, phi=10.0, einstein_radius=1.0)
            sie_2 = mass_profiles.EllipticalIsothermal(axis_ratio=0.6, phi=30.0, einstein_radius=1.2)

            deflection_angles_0 = sie_1.deflections_from_grid(np.array([[1.05, -0.55]]))
            deflection_angles_1 = sie_2.deflections_from_grid(np.array([[1.05, -0.55]]))

            deflection_angles = deflection_angles_0 + deflection_angles_1

            galaxy_sie = galaxy.Galaxy(redshift=0.5,
                                       mass_profile_1=mass_profiles.EllipticalIsothermal(axis_ratio=0.8, phi=10.0,
                                                                                         einstein_radius=1.0),
                                       mass_profile_2=mass_profiles.EllipticalIsothermal(axis_ratio=0.6, phi=30.0,
                                                                                         einstein_radius=1.2))

            galaxy_deflection_angles = galaxy_sie.deflections_from_grid(np.array([[1.05, -0.55]]))

            assert deflection_angles[0,0] == galaxy_deflection_angles[0,0]
            assert deflection_angles[0,1] == galaxy_deflection_angles[0,1]

        def test__three_profile_galaxy__deflection_angles_is_sum_of_individual_profiles(self, sie_1, sie_2, sie_3,
                                                                                        galaxy_sie_3):

            deflection_angles_0 = sie_1.deflections_from_grid(np.array([[1.05, -0.55]]))
            deflection_angles_1 = sie_2.deflections_from_grid(np.array([[1.05, -0.55]]))
            deflection_angles_2 = sie_3.deflections_from_grid(np.array([[1.05, -0.55]]))

            deflection_angles = deflection_angles_0 + deflection_angles_1 + deflection_angles_2

            galaxy_deflection_angles = galaxy_sie_3.deflections_from_grid(np.array([[1.05, -0.55]]))

            assert deflection_angles[0,0] == galaxy_deflection_angles[0,0]
            assert deflection_angles[0,1] == galaxy_deflection_angles[0,1]

        def test__three_profile_galaxy__individual_deflection_angles_can_be_extracted(self, sie_1, sie_2, sie_3,
                                                                                      galaxy_sie_3):

            deflection_angles_0 = sie_1.deflections_from_grid(np.array([[1.05, -0.55]]))
            deflection_angles_1 = sie_2.deflections_from_grid(np.array([[1.05, -0.55]]))
            deflection_angles_2 = sie_3.deflections_from_grid(np.array([[1.05, -0.55]]))

            galaxy_deflection_angles = galaxy_sie_3.deflections_from_grid_individual(np.array([[1.05, -0.55]]))

            assert (deflection_angles_0 == galaxy_deflection_angles[0]).all()
            assert (deflection_angles_1 == galaxy_deflection_angles[1]).all()
            assert (deflection_angles_2 == galaxy_deflection_angles[2]).all()

    class TestDimensionlessMassWithinCircle:

        def test__one_profile_galaxy__integral_is_same_as_individual_profile(self, sie_1, galaxy_sie_1):
            integral_radius = 5.5

            mass_integral = sie_1.dimensionless_mass_within_circle(radius=integral_radius)

            galaxy_mass_integral = galaxy_sie_1.dimensionless_mass_within_circles(radius=integral_radius)

            assert mass_integral == galaxy_mass_integral

        def test__two_profile_galaxy__integral_is_sum_of_individual_profiles(self):
            sie_1 = mass_profiles.EllipticalIsothermal(axis_ratio=0.8, phi=10.0, einstein_radius=1.0)
            sie_2 = mass_profiles.EllipticalIsothermal(axis_ratio=0.6, phi=30.0, einstein_radius=1.2)

            integral_radius = 5.5

            mass_integral = sie_1.dimensionless_mass_within_circle(radius=integral_radius)
            mass_integral += sie_2.dimensionless_mass_within_circle(radius=integral_radius)

            galaxy_sie = galaxy.Galaxy(redshift=0.5,
                                       mass_profile_1=mass_profiles.EllipticalIsothermal(axis_ratio=0.8, phi=10.0,
                                                                                         einstein_radius=1.0),
                                       mass_profile_2=mass_profiles.EllipticalIsothermal(axis_ratio=0.6, phi=30.0,
                                                                                         einstein_radius=1.2))

            galaxy_mass_integral = galaxy_sie.dimensionless_mass_within_circles(radius=integral_radius)

            assert mass_integral == galaxy_mass_integral

        def test__three_profile_galaxy__integral_is_sum_of_individual_profiles(self, sie_1, sie_2, sie_3,
                                                                               galaxy_sie_3):
            integral_radius = 5.5

            mass_integral = sie_1.dimensionless_mass_within_circle(radius=integral_radius)
            mass_integral += sie_2.dimensionless_mass_within_circle(radius=integral_radius)
            mass_integral += sie_3.dimensionless_mass_within_circle(radius=integral_radius)

            galaxy_mass_integral = galaxy_sie_3.dimensionless_mass_within_circles(radius=integral_radius)

            assert mass_integral == galaxy_mass_integral

        def test__three_profile_galaxy__individual_integrals_can_be_extracted(self, sie_1, sie_2, sie_3,
                                                                              galaxy_sie_3):
            integral_radius = 5.5

            mass_integral_1 = sie_1.dimensionless_mass_within_circle(radius=integral_radius)
            mass_integral_2 = sie_2.dimensionless_mass_within_circle(radius=integral_radius)
            mass_integral_3 = sie_3.dimensionless_mass_within_circle(radius=integral_radius)

            galaxy_mass_integral = galaxy_sie_3.dimensionless_mass_within_circles_individual(
                radius=integral_radius)

            assert mass_integral_1 == galaxy_mass_integral[0]
            assert mass_integral_2 == galaxy_mass_integral[1]
            assert mass_integral_3 == galaxy_mass_integral[2]

    class TestDimensionlessMassWithinEllipse:

        def test__one_profile_galaxy__integral_is_same_as_individual_profile(self, sie_1, galaxy_sie_1):
            integral_radius = 0.5

            dimensionless_mass_integral = sie_1.dimensionless_mass_within_ellipse(major_axis=integral_radius)

            galaxy_dimensionless_mass_integral = galaxy_sie_1.dimensionless_mass_within_ellipses(
                major_axis=integral_radius)

            assert dimensionless_mass_integral == galaxy_dimensionless_mass_integral

        def test__two_profile_galaxy__integral_is_sum_of_individual_profiles(self, sie_1, sie_2, galaxy_sie_2):
            integral_radius = 5.5

            dimensionless_mass_integral = sie_1.dimensionless_mass_within_ellipse(major_axis=integral_radius)
            dimensionless_mass_integral += sie_2.dimensionless_mass_within_ellipse(major_axis=integral_radius)

            galaxy_dimensionless_mass_integral = galaxy_sie_2.dimensionless_mass_within_ellipses(
                major_axis=integral_radius)

            assert dimensionless_mass_integral == galaxy_dimensionless_mass_integral

        def test__three_profile_galaxy__integral_is_sum_of_individual_profiles(self, sie_1, sie_2, sie_3,
                                                                               galaxy_sie_3):
            integral_radius = 5.5

            dimensionless_mass_integral = sie_1.dimensionless_mass_within_ellipse(major_axis=integral_radius)
            dimensionless_mass_integral += sie_2.dimensionless_mass_within_ellipse(major_axis=integral_radius)
            dimensionless_mass_integral += sie_3.dimensionless_mass_within_ellipse(major_axis=integral_radius)

            galaxy_dimensionless_mass_integral = galaxy_sie_3.dimensionless_mass_within_ellipses(
                major_axis=integral_radius)

            assert dimensionless_mass_integral == galaxy_dimensionless_mass_integral

        def test__three_profile_galaxy__individual_integrals_can_be_extracted(self, sie_1, sie_2, sie_3,
                                                                              galaxy_sie_3):
            integral_radius = 5.5

            dimensionless_mass_integral_1 = sie_1.dimensionless_mass_within_ellipse(major_axis=integral_radius)
            dimensionless_mass_integral_2 = sie_2.dimensionless_mass_within_ellipse(major_axis=integral_radius)
            dimensionless_mass_integral_3 = sie_3.dimensionless_mass_within_ellipse(major_axis=integral_radius)

            galaxy_dimensionless_mass_integral = galaxy_sie_3.dimensionless_mass_within_ellipses_individual(
                major_axis=integral_radius)

            assert dimensionless_mass_integral_1 == galaxy_dimensionless_mass_integral[0]
            assert dimensionless_mass_integral_2 == galaxy_dimensionless_mass_integral[1]
            assert dimensionless_mass_integral_3 == galaxy_dimensionless_mass_integral[2]

    class TestSymmetricProfiles:

        def test_1d_symmetry(self):

            isothermal_1 = mass_profiles.EllipticalIsothermal(axis_ratio=0.5, phi=45.0,
                                                              einstein_radius=1.0)

            isothermal_2 = mass_profiles.EllipticalIsothermal(centre=(100, 0), axis_ratio=0.5, phi=45.0,
                                                              einstein_radius=1.0)

            galaxy_isothermal = galaxy.Galaxy(redshift=0.5, mass_profile_1=isothermal_1, mass_profile_2=isothermal_2)

            assert galaxy_isothermal.surface_density_from_grid(
                np.array([[1.0, 0.0]])) == galaxy_isothermal.surface_density_from_grid(np.array([[99.0, 0.0]]))

            assert galaxy_isothermal.surface_density_from_grid(
                np.array([[49.0, 0.0]])) == galaxy_isothermal.surface_density_from_grid(np.array([[51.0, 0.0]]))

            assert galaxy_isothermal.potential_from_grid(np.array([[1.0, 0.0]])) == pytest.approx(
                galaxy_isothermal.potential_from_grid(np.array([[99.0, 0.0]])), 1e-6)

            assert galaxy_isothermal.potential_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                galaxy_isothermal.potential_from_grid(np.array([[51.0, 0.0]])), 1e-6)

            assert galaxy_isothermal.deflections_from_grid(np.array([[1.0, 0.0]])) == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[99.0, 0.0]])), 1e-6)

            assert galaxy_isothermal.deflections_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[51.0, 0.0]])), 1e-6)

        def test_2d_symmetry(self):

            isothermal_1 = mass_profiles.SphericalIsothermal(einstein_radius=1.0)

            isothermal_2 = mass_profiles.SphericalIsothermal(centre=(100, 0), einstein_radius=1.0)

            isothermal_3 = mass_profiles.SphericalIsothermal(centre=(0, 100), einstein_radius=1.0)

            isothermal_4 = mass_profiles.SphericalIsothermal(centre=(100, 100), einstein_radius=1.0)

            galaxy_isothermal = galaxy.Galaxy(redshift=0.5,
                                              mass_profile_1=isothermal_1, mass_profile_2=isothermal_2,
                                              mass_profile_3=isothermal_3, mass_profile_4=isothermal_4)

            assert galaxy_isothermal.surface_density_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                galaxy_isothermal.surface_density_from_grid(np.array([[51.0, 0.0]])), 1e-5)

            assert galaxy_isothermal.surface_density_from_grid(np.array([[0.0, 49.0]])) == pytest.approx(
                galaxy_isothermal.surface_density_from_grid(np.array([[0.0, 51.0]])), 1e-5)

            assert galaxy_isothermal.surface_density_from_grid(np.array([[100.0, 49.0]])) == pytest.approx(
                galaxy_isothermal.surface_density_from_grid(np.array([[100.0, 51.0]])), 1e-5)

            assert galaxy_isothermal.surface_density_from_grid(np.array([[49.0, 49.0]])) == pytest.approx(
                galaxy_isothermal.surface_density_from_grid(np.array([[51.0, 51.0]])), 1e-5)

            assert galaxy_isothermal.potential_from_grid(np.array([[49.0, 0.0]])) == pytest.approx(
                galaxy_isothermal.potential_from_grid(np.array([[51.0, 0.0]])), 1e-5)

            assert galaxy_isothermal.potential_from_grid(np.array([[0.0, 49.0]])) == pytest.approx(
                galaxy_isothermal.potential_from_grid(np.array([[0.0, 51.0]])), 1e-5)

            assert galaxy_isothermal.potential_from_grid(np.array([[100.0, 49.0]])) == pytest.approx(
                galaxy_isothermal.potential_from_grid(np.array([[100.0, 51.0]])), 1e-5)

            assert galaxy_isothermal.potential_from_grid(np.array([[49.0, 49.0]])) == pytest.approx(
                galaxy_isothermal.potential_from_grid(np.array([[51.0, 51.0]])), 1e-5)

            assert -1.0 * galaxy_isothermal.deflections_from_grid(np.array([[49.0, 0.0]]))[0, 0] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[51.0, 0.0]]))[0, 0], 1e-5)

            assert 1.0 * galaxy_isothermal.deflections_from_grid(np.array([[0.0, 49.0]]))[0, 0] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[0.0, 51.0]]))[0, 0], 1e-5)

            assert 1.0 * galaxy_isothermal.deflections_from_grid(np.array([[100.0, 49.0]]))[0, 0] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[100.0, 51.0]]))[0, 0], 1e-5)

            assert -1.0 * galaxy_isothermal.deflections_from_grid(np.array([[49.0, 49.0]]))[0, 0] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[51.0, 51.0]]))[0, 0], 1e-5)

            assert 1.0 * galaxy_isothermal.deflections_from_grid(np.array([[49.0, 0.0]]))[0, 1] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[51.0, 0.0]]))[0, 1], 1e-5)

            assert -1.0 * galaxy_isothermal.deflections_from_grid(np.array([[0.0, 49.0]]))[0, 1] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[0.0, 51.0]]))[0, 1], 1e-5)

            assert -1.0 * galaxy_isothermal.deflections_from_grid(np.array([[100.0, 49.0]]))[0, 1] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[100.0, 51.0]]))[0, 1], 1e-5)

            assert -1.0 * galaxy_isothermal.deflections_from_grid(np.array([[49.0, 49.0]]))[0, 1] == pytest.approx(
                galaxy_isothermal.deflections_from_grid(np.array([[51.0, 51.0]]))[0, 1], 1e-5)


class TestHyperGalaxy(object):

    class TestContributionMaps(object):

        def test__model_image_all_1s__factor_is_0__contributions_all_1s(self):
            galaxy_image = np.ones((3,))

            hyp = galaxy.HyperGalaxy(contribution_factor=0.0)
            contributions = hyp.contributions_from_model_images(model_image=galaxy_image, galaxy_image=galaxy_image,
                                                                minimum_value=0.0)

            assert (contributions == np.ones((3,))).all()

        def test__different_values__factor_is_1__contributions_are_value_divided_by_factor_and_max(self):
            galaxy_image = np.array([0.5, 1.0, 1.5])

            hyp = galaxy.HyperGalaxy(contribution_factor=1.0)
            contributions = hyp.contributions_from_model_images(model_image=galaxy_image, galaxy_image=galaxy_image,
                                                                minimum_value=0.0)

            assert (contributions == np.array([(0.5 / 1.5) / (1.5 / 2.5), (1.0 / 2.0) / (1.5 / 2.5), 1.0])).all()

        def test__different_values__threshold_is_1_minimum_threshold_included__wipes_1st_value_to_0(self):
            galaxy_image = np.array([0.5, 1.0, 1.5])

            hyp = galaxy.HyperGalaxy(contribution_factor=1.0)
            contributions = hyp.contributions_from_model_images(model_image=galaxy_image, galaxy_image=galaxy_image,
                                                                minimum_value=0.6)

            assert (contributions == np.array([0.0, (1.0 / 2.0) / (1.5 / 2.5), 1.0])).all()

    class TestScaledNoise(object):

        def test__contribution_all_1s__noise_factor_2__noise_adds_double(self):
            noise = np.array([1.0, 2.0, 3.0])
            galaxy_contributions = np.ones((3, 1))

            hyp = galaxy.HyperGalaxy(contribution_factor=0.0, noise_factor=2.0, noise_power=1.0)

            scaled_noise = hyp.scaled_noise_for_contributions(noise=noise, contributions=galaxy_contributions)

            assert (scaled_noise == np.array([2.0, 4.0, 6.0])).all()

        def test__same_as_above_but_contributions_vary(self):
            noise = np.array([1.0, 2.0, 3.0])
            galaxy_contributions = np.array([[0.0, 0.5, 1.0]])

            hyp = galaxy.HyperGalaxy(contribution_factor=0.0, noise_factor=2.0, noise_power=1.0)

            scaled_noise = hyp.scaled_noise_for_contributions(noise=noise, contributions=galaxy_contributions)

            assert (scaled_noise == np.array([0.0, 2.0, 6.0])).all()

        def test__same_as_above_but_change_noise_scale_terms(self):
            noise = np.array([1.0, 2.0, 3.0])
            galaxy_contributions = np.array([[0.0, 0.5, 1.0]])

            hyp = galaxy.HyperGalaxy(contribution_factor=0.0, noise_factor=2.0, noise_power=2.0)

            scaled_noise = hyp.scaled_noise_for_contributions(noise=noise, contributions=galaxy_contributions)

            assert (scaled_noise == np.array([0.0, 2.0, 18.0])).all()


class TestBooleanProperties(object):
    def test_has_pixelization(self):
        assert galaxy.Galaxy().has_pixelization is False
        assert galaxy.Galaxy(pixelization=object()).has_pixelization is True

    def test_has_hyper_galaxy(self):
        assert galaxy.Galaxy().has_pixelization is False
        assert galaxy.Galaxy(hyper_galaxy=object()).has_hyper_galaxy is True

    def test_has_profile(self):
        assert galaxy.Galaxy().has_profile is False
        assert galaxy.Galaxy(light_profile=light_profiles.LightProfile()).has_profile is True
        assert galaxy.Galaxy(mass_profile=mass_profiles.MassProfile()).has_profile is True



from autolens.analysis import analysis as an
from autolens.imaging import masked_image as mi
from autolens.analysis import galaxy_prior
from autolens.profiles import light_profiles, mass_profiles
from autolens.autopipe import model_mapper
from autolens.autopipe import non_linear
from autolens.pipeline import phase as ph
from autolens.pixelization import pixelization
from autolens.analysis import galaxy

import os
import logging

logging.basicConfig()
logger = logging.getLogger(__name__)


def source_only_pipeline(image, mask):
    """
    Fit a source-only image (i.e. no lens light component)

    Parameters
    ----------
    image: Image
        An image of a lens galaxy including metadata such as PSF, background noise and effective exposure time
    mask: Mask
        A mask describing which parts of the image should be excluded from the analysis

    Returns
    -------
    result: [Result]
        An array of results, one for each stage of the analysis
    """

    logger.info(
        """
        Pipeline 1:

        PURPOSE - Fit a source-only image (i.e. no lens light component)

        PREPROCESSING:

        - Mark the brightest regions / multiple images of the source.
        - Draw a circle tracing the source (Einstein Radius / centre)
        - Draw circle / ellipse for the mask.

        NOTES:

        Image: Observed image used throughout.
        Mask: Assume a large mask (e.g. 2") throughout - this value could be chosen in preprocessing.
        """
    )

    # Create an array in which to store results
    results = []

    # Create an analysis object for the image and mask
    masked_image = mi.MaskedImage(image, mask)

    logger.info(
        """
        1) Mass: SIE+Shear
           Source: Sersic
           NLO: LM
        """
    )

    # Create an optimizer
    phase1 = ph.ParameterizedPhase(optimizer=non_linear.MultiNest())

    # Define galaxy priors
    source_galaxy_prior = galaxy_prior.GalaxyPrior(light_profile=light_profiles.EllipticalSersic)
    lens_galaxy_prior = galaxy_prior.GalaxyPrior(spherical_mass_profile=mass_profiles.EllipticalIsothermal,
                                                 shear_mass_profile=mass_profiles.ExternalShear)

    # Add the galaxy priors to the optimizer
    phase1.add_variable_lens_galaxies([lens_galaxy_prior])
    phase1.add_variable_source_galaxies([source_galaxy_prior])

    # Analyse the system
    result_1 = phase1.fit(masked_image)

    # Add the result of the first analysis to the list
    results.append(result_1)

    logger.info(
        """
        2) Mass: SIE+Shear (priors from phase 1)
           Source: 'smooth' pixelization (include regularization parameter(s) in the model)
           NLO: LM
        """
    )
    # Create an optimizer
    phase = ph.ParameterizedPhase(optimizer=non_linear.MultiNest())
    optimizer_2 = non_linear.DownhillSimplex()

    # Define galaxy priors
    lens_galaxy_prior = galaxy_prior.GalaxyPrior(spherical_mass_profile=mass_profiles.EllipticalIsothermal,
                                                 shear_mass_profile=mass_profiles.ExternalShear)
    # The source galaxy is now represented by a pixelization
    source_galaxy_prior = galaxy_prior.GalaxyPrior(pixelization=pixelization.Rectangular)

    # Add the galaxy priors to the optimizer
    optimizer_2.variable.lens_galaxies = [lens_galaxy_prior]
    optimizer_2.variable.source_galaxies = [source_galaxy_prior]

    # Associate priors found in the first analysis with the new galaxy priors
    lens_galaxy_prior_result = result_1.priors.lens_galaxies[0]
    lens_galaxy_prior.spherical_mass_profile = lens_galaxy_prior_result.spherical_mass_profile
    lens_galaxy_prior.shear_mass_profile = lens_galaxy_prior_result.shear_mass_profile

    # Analyse the system
    result_2 = optimizer_2.fit(analysis)

    # Add the result of the second analysis to the list
    results.append(result_2)

    logger.info(
        """
        2H) Hyper-parameters: All included in model (most priors broad and uniform, but use previous phase 
            regularization as well)
            Mass: SIE+Shear (Fixed to highest likelihood model from phase 2)
            Source: 'noisy' pixelization
            NLO: MN
        """
    )
    # Create an optimizer
    optimizer_2h = non_linear.MultiNest()

    # Define a single galaxy prior that is a pixelized galaxy
    source_galaxy_prior = galaxy_prior.GalaxyPrior(pixelization=pixelization.Voronoi)

    #  Add the variable pixelization to the optimizer
    optimizer_2h.variable.source_galaxies = [source_galaxy_prior]

    # Set the regularization prior using results from analysis 2
    source_galaxy_prior.pixelization.regularization = result_2.priors.pixelization.regularization

    # Set a constant lens galaxy from the previous result
    optimizer_2h.constant.lens_galaxies = result_2.instance.lens_galaxies

    # Analyse the system
    result_2h = optimizer_2h.fit(analysis)

    # Add the result of analysis 2h to the results
    results.append(result_2h)

    logger.info(
        """
        a) Mass: SPLE+Shear (priors from Init phase 2)
           Source: 'noisy' pixelization (Fixed to init 2H hyper-parameters)
        """
    )
    # Create an optimizer
    optimizer_a = non_linear.MultiNest()

    # Define a lens galaxy prior
    lens_galaxy_prior = galaxy_prior.GalaxyPrior(spherical_power_law_mass_profile=mass_profiles.SphericalPowerLaw,
                                                 shear_mass_profile=mass_profiles.ExternalShear)

    # Add the lens galaxy prior to the optimizer
    optimizer_a.variable.lens_galaxies = [lens_galaxy_prior]

    # Set some lens galaxy priors using results from analysis 2
    lens_galaxy_prior_result = result_2.priors.lens_galaxies[0]
    lens_galaxy_prior.shear_mass_profile = lens_galaxy_prior_result.shear_mass_profile
    lens_galaxy_prior.spherical_power_law_mass_profile.centre = lens_galaxy_prior_result.spherical_mass_profile.centre

    # Set a source galaxy with a fixed pixelization from analysis 2
    optimizer_a.constant.source_galaxies = result_2h.instance.source_galaxies

    # Analyse the system
    result_a = optimizer_a.fit(analysis)

    # Add the result of the main analysis to the results
    results.append(result_a)

    logger.info("Pipeline complete")

    # Return the results
    return results


def lens_and_source_pipeline(image, lens_mask, source_mask, combined_mask):
    """
    Pipeline 2:

    PURPOSE - Fit a lens light + source image (mass model does not decomposed the light and dark matter)

    PREPROCESSING:

    - Mark the brightest regions / multiple images of the source.
    - Draw a circle around the source (Einstein Radius / center)
    - Mark the centre of the lens light.
    - Draw circle / ellipse for mask containing the lens galaxy.
    - Draw annulus for mask containing the source (excludes central regions of image where lens is).
    """

    # Create an array in which to store results
    results = []
    """

    1) Image: Observed image, includes lens+source.
       Mask: Circle / Ellipse containing entire lens galaxy.
       Light: Sersic (This phase simply subtracts the lens light from the image - the source is present and disrupts the 
       NLO: LM        fit but we choose not to care)
       Purpose: Provides lens subtracted image for subsequent phases.

    """
    # Create an analysis object for the image and a mask that means only lens light is visible
    lens_light_analysis = an.Analysis(image, lens_mask)

    optimizer_1 = non_linear.DownhillSimplex()

    lens_galaxy = galaxy_prior.GalaxyPrior(light_profile=light_profiles.EllipticalSersic)
    lens_galaxy.redshift = model_mapper.Constant(1)

    optimizer_1.variable.lens_galaxies = [lens_galaxy]

    result_1 = optimizer_1.fit(lens_light_analysis)

    results.append(result_1)

    # TODO: simplify this stage of the process
    light_grid = result_1.lens_galaxies[0].light_profile.intensity_at_coordinates(
        lens_light_analysis.coords_collection.image)

    source_image = image - lens_light_analysis.mapper_collection.data_to_pixel.map_to_2d(light_grid)

    """
    2) Image: The lens light subtracted image from phase 1.
       Mask: Annulus mask containing just source
       Light: None
       Mass: SIE+Shear
       Source: Sersic
       NLO: LM
       Purpose: Provides mass model priors for next phase.
    """

    source_analysis = an.Analysis(source_image, source_mask)

    optimizer_2 = non_linear.DownhillSimplex()

    lens_galaxy = galaxy_prior.GalaxyPrior(sie_mass_profile=mass_profiles.SphericalIsothermal,
                                           shear_mass_profile=mass_profiles.ExternalShear)
    source_galaxy = galaxy_prior.GalaxyPrior(light_profile=light_profiles.EllipticalSersic)

    optimizer_2.variable.lens_galaxies = [lens_galaxy]
    optimizer_2.variable.source_galaxies = [source_galaxy]

    result_2 = optimizer_2.fit(source_analysis)

    results.append(result_2)

    """
    3) Image: The lens light subtracted image from phase 1.
       Mask: Circle / Ellipse containing entire lens galaxy.
       Light: None
       Mass: SIE+Shear (priors from phase 2)
       Source: 'smooth' pixelization (include regularization parameter(s) in the model)
       NLO: LM
       Purpose: Refines mass model and sets up the source-plane pixelization regularization.
    """

    pixelized_source_analysis = an.Analysis(source_image, combined_mask)

    optimizer_3 = non_linear.DownhillSimplex()

    source_galaxy = galaxy_prior.GalaxyPrior(pixelization=pixelization.SquarePixelization)
    lens_galaxy = galaxy_prior.GalaxyPrior(sie_mass_profile=mass_profiles.SphericalIsothermal,
                                           shear_mass_profile=mass_profiles.ExternalShear)

    lens_galaxy_result = result_2.priors.lens_galaxies[0]

    lens_galaxy.sie_mass_profile.centre = lens_galaxy_result.sie_mass_profile.centre
    lens_galaxy.sie_mass_profile.einstein_radius = lens_galaxy_result.sie_mass_profile.einstein_radius

    lens_galaxy.shear_mass_profile.magnitude = lens_galaxy_result.shear_mass_profile.magnitude
    lens_galaxy.shear_mass_profile.phi = lens_galaxy_result.shear_mass_profile.phi

    optimizer_3.variable.lens_galaxies = [lens_galaxy]
    optimizer_3.variable.source_galaxies = [source_galaxy]

    result_3 = optimizer_3.fit(pixelized_source_analysis)
    results.append(result_3)

    """
    4) Image: Observed image, includes lens+source.
       Mask: Circle / Ellipse containing entire lens galaxy.
       Light: Sersic + Exponential (shared centre / phi, include Instrumentation lens light noise scaling parameters in 
              model)
       Mass: SIE+Shear (fixed to results from phase 3)
       Source: 'smooth' pixelization (include regularization parameter(s) in the model, using previous phase prior)
       NLO: LM
       Purpose: To fit a complex light profile, we need to do so simultaneously with the source reconstruction to avoid
                systematics. Thus, this rather confusing phase sets it up so that the mass profile is fixed whilst we
                get ourselves a good multi-component light profile.
    """

    full_light_analysis = an.Analysis(image, combined_mask)

    optimizer_4 = non_linear.DownhillSimplex(include_hyper_image=True)

    lens_galaxy = galaxy_prior.GalaxyPrior(sie_mass_profile=result_3.instance.sie_mass_profile,
                                           shear_mass_profile=result_3.instance.shear_mass_profile,
                                           sersic_light_profile=light_profiles.EllipticalSersic,
                                           exponential_light_profile=light_profiles.EllipticalExponential,
                                           hyper_galaxy=galaxy.HyperGalaxy)

    lens_galaxy.sersic_light_profile.centre = lens_galaxy.exponential_light_profile.centre
    lens_galaxy.sersic_light_profile.phi = lens_galaxy.exponential_light_profile.phi

    source_galaxy = galaxy_prior.GalaxyPrior(pixelization=pixelization.SquarePixelization)

    pixelization_result = result_3.priors.pixelization
    source_galaxy.pixelization.pixels = pixelization_result.pixels
    source_galaxy.pixelization.regularization_coefficients = pixelization_result.regularization_coefficients

    optimizer_4.variable.lens_galaxies = [lens_galaxy]
    optimizer_4.variable.source_galaxies = [source_galaxy]

    optimizer_4.fit(full_light_analysis)


default_config_path = '{}/../../../config/'.format(os.path.dirname(os.path.realpath(__file__)))
default_data_path = '{}/../../weighted_data/'.format(os.path.dirname(os.path.realpath(__file__)))
# default_results_path = '{}/../../../results/'.format(os.path.dirname(os.path.realpath(__file__)))
default_results_path = '/home/jammy/results/'


class PipelinePaths(object):

    def __init__(self, data_name, config_path=default_config_path, data_path=default_data_path,
                 results_path=default_results_path):
        self.data_name = data_name
        self.config_path = config_path
        self.data_path = data_path + data_name + '/'
        self.results_path = results_path


def profiles_pipeline(paths, image, mask):
    """
     Fit an image with profiles, gradually adding extra profiles.

     Parameters
     ----------
     paths
     image: Image
         An image of a lens galaxy including metadata such as PSF, background noise and effective exposure time
     mask: Mask
         A mask describing which parts of the image should be excluded from the analysis

     Returns
     -------
     result: [Result]
         An array of results, one for each stage of the analysis
     """

    logger.info(
        """
        Pipeline Profiles:

        PURPOSE - Fit profiles to an image.

        PREPROCESSING:

        - N/A

        NOTES:

        Image: Observed image used throughout.
        Mask: Assume a large mask (e.g. 2.8") throughout - this value could be chosen in preprocessing.
        """
    )

    pipeline_name = 'Hyper_Lens'
    pipeline_path = paths.results_path + pipeline_name + '/'
    pipeline_data_path = pipeline_path + paths.data_name + '/'
    if not os.path.exists(pipeline_path): os.mkdir(pipeline_path)
    if not os.path.exists(pipeline_data_path): os.mkdir(pipeline_data_path)

    # Create an array in which to store results
    results = []

    # Create an analysis object for the image and mask
    analysis = an.Analysis(image, mask)

    logger.info(
        """
        1) Light: Elliptical Sersic x1
        """
    )

    analysis_path = pipeline_data_path + '1_lens_init/'

    # Create an optimizer
    optimizer_1 = non_linear.DownhillSimplex(path=analysis_path)

    # Define galaxy priors
    lens_galaxy_prior = galaxy_prior.GalaxyPrior(sersic_light_profile=light_profiles.EllipticalSersic)
    lens_galaxy_prior.sersic_light_profile.centre = model_mapper.Constant((0.0, 0.0))
    lens_galaxy_prior.sersic_light_profile.axis_ratio = model_mapper.Constant(0.8)
    lens_galaxy_prior.sersic_light_profile.phi = model_mapper.Constant(90.0)
    lens_galaxy_prior.sersic_light_profile.intensity = model_mapper.Constant(0.5)
    lens_galaxy_prior.sersic_light_profile.effective_radius = model_mapper.Constant(1.3)

    lens_galaxy_prior.redshift = model_mapper.Constant(1.0)

    # Add the galaxy priors to the optimizer
    optimizer_1.variable.lens_galaxies = [lens_galaxy_prior]

    # Analyse the system
    result_1 = optimizer_1.fit(analysis)

    # Add the result of the second analysis to the list
    results.append(result_1)

    logger.info(
        """
        Pipeline Profiles:

        PURPOSE - Fit profiles to an image.

        PREPROCESSING:

        - N/A

        NOTES:

        Image: Observed image used throughout.
        Mask: Assume a large mask (e.g. 2.8") throughout - this value could be chosen in preprocessing.
        """
    )

    # Create an array in which to store results
    results = []

    # TODO : These images should come from the results of the previous phase

    import numpy as np

    model_image = np.ones((mask.pixels_in_mask,))
    galaxy_images = [np.ones((mask.pixels_in_mask,))]

    # TODO : minimum value hsould be 0 if the galaxy was a profile, and 0.02 if it was a pixelization.

    minimum_values = [0.0]

    # Create an analysis object for the image and mask
    analysis = an.Analysis(image, mask, model_image=model_image, galaxy_images=galaxy_images,
                           minimum_values=minimum_values)

    logger.info(
        """
        2) Light: Elliptical Sersic x2 (The results of the first Sersic are fixed)
        """
    )

    analysis_path = pipeline_data_path + '1h_lens/'

    # Create an optimizer
    optimizer_1 = non_linear.DownhillSimplex(path=analysis_path)

    # Define galaxy priors
    lens_galaxy_prior = galaxy_prior.GalaxyPrior(
        sersic_light_profile=result_1.instance.lens_galaxies[0].sersic_light_profile,
        hyper_galaxy=galaxy.HyperGalaxy)

    # Add the galaxy priors to the optimizer
    optimizer_1.variable.lens_galaxies = [lens_galaxy_prior]

    # Analyse the system
    result_1 = optimizer_1.fit(analysis)

    # Add the result of the second analysis to the list
    results.append(result_1)


# def profiles_pipeline(paths, image, mask):
#     """
#      Fit an image with profiles, gradually adding extra profiles.
#
#      Parameters
#      ----------
#      image: Image
#          An image of a lens galaxy including metadata such as PSF, background noise and effective exposure time
#      mask: Mask
#          A mask describing which parts of the image should be excluded from the analysis
#
#      Returns
#      -------
#      result: [Result]
#          An array of results, one for each stage of the analysis
#      """
#
#     logger.info(
#         """
#         Pipeline Profiles:
#
#         PURPOSE - Fit profiles to an image.
#
#         PREPROCESSING:
#
#         - N/A
#
#         NOTES:
#
#         Image: Observed image used throughout.
#         Mask: Assume a large mask (e.g. 2.8") throughout - this value could be chosen in preprocessing.
#         """
#     )
#
#     pipeline_name = 'Profiles'
#     pipeline_path = paths.results_path+pipeline_name+'/'
#     pipeline_data_path = pipeline_path+paths.data_name+'/'
#     if not os.path.exists(pipeline_path): os.mkdir(pipeline_path)
#     if not os.path.exists(pipeline_data_path): os.mkdir(pipeline_data_path)
#
#     # Create an array in which to store results
#     results = []
#
#     # Create an analysis object for the image and mask
#     analysis = an.Analysis(image, mask)
#
#     logger.info(
#         """
#         1) Light: Elliptical Sersic x1
#         """
#     )
#
#     analysis_path = pipeline_data_path + '1_profile_1st/'
#
#     # Create an optimizer
#     optimizer_1 = non_linear.DownhillSimplex(path=analysis_path)
#
#     # Define galaxy priors
#     lens_galaxy_prior = galaxy_prior.GalaxyPrior(sersic_light_profile=light_profiles.EllipticalSersic)
#     lens_galaxy_prior.sersic_light_profile.centre = model_mapper.Constant((1.0, 1.0))
#     # TODO : Can we get rid of the x2 redshift?
#     lens_galaxy_prior.redshift.redshift = model_mapper.Constant(1.0)
#
#     # Add the galaxy priors to the optimizer
#     print(lens_galaxy_prior)
#     stop
#     optimizer_1.lens_galaxies = [lens_galaxy_prior]
#
#     # Analyse the system
#     result_1 = optimizer_1.fit(analysis)
#
#     # Add the result of the second analysis to the list
#     results.append(result_1)
#
#     logger.info(
#         """
#         Pipeline Profiles:
#
#         PURPOSE - Fit profiles to an image.
#
#         PREPROCESSING:
#
#         - N/A
#
#         NOTES:
#
#         Image: Observed image used throughout.
#         Mask: Assume a large mask (e.g. 2.8") throughout - this value could be chosen in preprocessing.
#         """
#     )
#
#     pipeline_name = 'Profiles'
#     pipeline_path = paths.results_path+pipeline_name+'/'
#     pipeline_data_path = pipeline_path+paths.data_name+'/'
#     if not os.path.exists(pipeline_path): os.mkdir(pipeline_path)
#     if not os.path.exists(pipeline_data_path): os.mkdir(pipeline_data_path)
#
#     # Create an array in which to store results
#     results = []
#
#     # Create an analysis object for the image and mask
#     analysis = an.Analysis(image, mask)
#
#     logger.info(
#         """
#         2) Light: Elliptical Sersic x2 (The results of the first Sersic are fixed)
#         """
#     )
#
#     analysis_path = pipeline_data_path + '1_profile_1st/'
#
#     # Create an optimizer
#     optimizer_1 = non_linear.DownhillSimplex(path=analysis_path)
#
#     # Define galaxy priors
#     lens_galaxy_prior = galaxy_prior.GalaxyPrior(sersic_light_profile=light_profiles.EllipticalSersic)
#     lens_galaxy_prior.sersic_light_profile.centre.centre_0 = model_mapper.GaussianPrior(1.0, 1.0)
#     lens_galaxy_prior.sersic_light_profile.centre.centre_1 = model_mapper.GaussianPrior(1.0, 1.0)
#
#     # Add the galaxy priors to the optimizer
#     optimizer_1.lens_galaxies = [lens_galaxy_prior]
#
#     # Analyse the system
#     result_1 = optimizer_1.fit(analysis)
#
#     # Add the result of the second analysis to the list
#     results.append(result_1)

"""

4H) Hyper-parameters: All included in model (most priors broad and uniform, but use previous phase regularization as well)
    Image: The lens light subtracted image from phase 1.
    Mask: Circle / Ellipse containing entire lens galaxy.
    Mass: SIE+Shear (Fixed to highest likelihood model from phase 2)
    Source: 'noisy' pixelization
    NLO: MN.
    """



import getdist

from autolens import exc
import math
import os
import pymultinest
import scipy.optimize
import numpy as np
from autolens.imaging import hyper_image
from autolens import config
from autolens.autopipe import model_mapper as mm
import logging

logging.basicConfig()
logger = logging.getLogger(__name__)

default_path = '{}/../../output'.format(os.path.dirname(os.path.realpath(__file__)))
dir_path = os.path.dirname(os.path.realpath(__file__))

SIMPLEX_TUPLE_WIDTH = 0.1

def generate_parameter_latex(parameters, subscript=''):
    """Generate a latex label for a non-linear search parameter.

    This is used for the param names file and outputting the nlo of a run to a latex table.

    Parameters
    ----------
    parameters : [str]
        The parameter names to be converted to latex.
    subscript : str
        The subscript of the latex entry, often giving the parameter type (e.g. light or dark matter) or numerical \
        number of the component of the model_mapper.

    """

    latex = []

    if subscript == '':
        for param in parameters:
            latex.append('$' + param + '$')
    else:
        for param in parameters:
            latex.append('$' + param + r'_{\mathrm{' + subscript + '}}$')

    return latex


class Result(object):

    def __init__(self, constant, likelihood, variable=None):
        """
        The result of an optimization.

        Parameters
        ----------
        constant: mm.ModelInstance
            An instance object comprising the class instances that gave the optimal fit
        likelihood: float
            A value indicating the likelihood given by the optimal fit
        variable: mm.ModelMapper
            An object comprising priors determined by this stage of the analysis
        """
        self.constant = constant
        self.likelihood = likelihood
        self.variable = variable

    def __str__(self):
        return "Analysis Result:\n{}".format(
            "\n".join(["{}: {}".format(key, value) for key, value in self.__dict__.items()]))


class NonLinearOptimizer(object):

    def __init__(self, include_hyper_image=False, model_mapper=None,
                 config_path=None, path=default_path, name=None, **classes):
        """Abstract base class for non-linear optimizers.

        This class sets up the file structure for the non-linear optimizer nlo, which are standardized across all \
        non-linear optimizers.

        Parameters
        ------------
        path : str
            The path where the non-linear analysis nlo are stored.
        obj_name : str
            Unique identifier of the weighted_data being analysed (e.g. the name of the weighted_data set)
        """
        self.nlo_config = config.NamedConfig(
            "{}/../../config/non_linear.ini".format(dir_path) if config_path is None else config_path,
            self.__class__.__name__)

        self.path = "{}/{}".format(path, name) if name is not None else path
        self.variable = mm.ModelMapper() if model_mapper is None else model_mapper
        self.constant = mm.ModelInstance()

        self.file_param_names = "{}/{}".format(self.path, '/multinest.paramnames')
        self.file_model_info = "{}/{}".format(self.path, '/model.info')

        # If the include_hyper_image flag is set to True make this an additional prior model
        if include_hyper_image:
            self.hyper_image = mm.PriorModel(hyper_image.HyperImage, config=self.variable.config)

    def save_model_info(self):
        if not os.path.exists(self.path):
            os.makedirs(self.path)  # Create results folder if doesnt exist

    #    self.create_param_names()
        self.variable.output_model_info(self.file_model_info)
        self.variable.check_model_info(self.file_model_info)

    def fit(self, analysis):
        raise NotImplementedError("Fitness function must be overridden by non linear optimizers")

    # def create_param_names(self):
    #     """The param_names file lists every parameter's name and Latex tag, and is used for *GetDist* visualization.
    #
    #     The parameter names are determined from the class instance names of the model_mapper. Latex tags are \
    #     properties of each model class."""
    #     param_names = open(self.file_param_names, 'w')
    #
    #     for prior_name, prior_model in self.variable.flat_prior_models:
    #
    #         param_labels = prior_model.cls.parameter_labels.__get__(prior_model.cls)
    #         component_number = prior_model.cls().component_number
    #         subscript = prior_model.cls.subscript.__get__(prior_model.cls) + str(component_number + 1)
    #
    #         param_labels = generate_parameter_latex(param_labels, subscript)
    #
    #         for param_no, param in enumerate(self.variable.class_priors_dict[prior_name]):
    #             line = prior_name + '_' + param[0]
    #             line += ' ' * (40 - len(line)) + param_labels[param_no]
    #
    #             param_names.write(line + '\n')
    #
    #     param_names.close()


class DownhillSimplex(NonLinearOptimizer):

    def __init__(self, include_hyper_image=False, model_mapper=None, path=default_path,
                 fmin=scipy.optimize.fmin, name=None):
        super(DownhillSimplex, self).__init__(include_hyper_image=include_hyper_image,
                                              model_mapper=model_mapper, path=path, name=name)

        self.xtol = self.nlo_config.get("xtol", float)
        self.ftol = self.nlo_config.get("ftol", float)
        self.maxiter = self.nlo_config.get("maxiter", int)
        self.maxfun = self.nlo_config.get("maxfun", int)

        self.full_output = self.nlo_config.get("full_output", int)
        self.disp = self.nlo_config.get("disp", int)
        self.retall = self.nlo_config.get("retall", int)

        self.fmin = fmin

        logger.debug("Creating DownhillSimplex NLO")

    def fit(self, analysis):
        initial_vector = self.variable.physical_vector_from_prior_medians

        class Fitness(object):
            def __init__(self, instance_from_physical_vector, constant):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.constant = constant

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)
                for key, value in self.constant.__dict__.items():
                    setattr(instance, key, value)

                likelihood = analysis.fit(**instance.__dict__)
                self.result = Result(instance, likelihood)

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(self.variable.instance_from_physical_vector, self.constant)

        logger.info("Running DownhillSimplex...")
        output = self.fmin(fitness_function, x0=initial_vector)
        logger.info("DownhillSimplex complete")
        res = fitness_function.result

        # Create a set of Gaussian priors from this result and associate them with the result object.
        res.variable = self.variable.mapper_from_gaussian_means(output)

        return res


class MultiNest(NonLinearOptimizer):

    def __init__(self, include_hyper_image=False, model_mapper=None, path=default_path,
                 sigma_limit=3, run=pymultinest.run, name=None):
        """Class to setup and run a MultiNest analysis and output the MultiNest nlo.

        This interfaces with an input model_mapper, which is used for setting up the individual model instances that \
        are passed to each iteration of MultiNest.

        Parameters
        ------------
        path : str
            The path where the non_linear nlo are stored.
        """

        super(MultiNest, self).__init__(include_hyper_image=include_hyper_image, model_mapper=model_mapper,
                                        path=path, name=name)

        self.file_summary = "{}/{}".format(self.path, 'summary.txt')
        self.file_weighted_samples = "{}/{}".format(self.path, 'multinest.txt')
        self._weighted_sample_model = None
        self.sigma_limit = sigma_limit

        self.importance_nested_sampling = self.nlo_config.get('importance_nested_sampling', bool)
        self.multimodal = self.nlo_config.get('multimodal', bool)
        self.const_efficiency_mode = self.nlo_config.get('const_efficiency_mode', bool)
        self.n_live_points = self.nlo_config.get('n_live_points', int)
        self.evidence_tolerance = self.nlo_config.get('evidence_tolerance', float)
        self.sampling_efficiency = self.nlo_config.get('sampling_efficiency', float)
        self.n_iter_before_update = self.nlo_config.get('n_iter_before_update', int)
        self.null_log_evidence = self.nlo_config.get('null_log_evidence', float)
        self.max_modes = self.nlo_config.get('max_modes', int)
        self.mode_tolerance = self.nlo_config.get('mode_tolerance', float)
        self.outputfiles_basename = self.nlo_config.get('outputfiles_basename', str)
        self.seed = self.nlo_config.get('seed', int)
        self.verbose = self.nlo_config.get('verbose', bool)
        self.resume = self.nlo_config.get('resume', bool)
        self.context = self.nlo_config.get('context', int)
        self.write_output = self.nlo_config.get('write_output', bool)
        self.log_zero = self.nlo_config.get('log_zero', float)
        self.max_iter = self.nlo_config.get('max_iter', int)
        self.init_MPI = self.nlo_config.get('init_MPI', bool)
        self.run = run

        logger.debug("Creating MultiNest NLO")

    @property
    def pdf(self):
        return getdist.mcsamples.loadMCSamples(self.file_weighted_samples)

    def fit(self, analysis):
        self.save_model_info()

        class Fitness(object):
            def __init__(self, instance_from_physical_vector, constant):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.constant = constant
                self.max_likelihood = -np.inf

            def __call__(self, cube, ndim, nparams, lnew):
                instance = self.instance_from_physical_vector(cube)
                for key, value in self.constant.__dict__.items():
                    setattr(instance, key, value)

                likelihood = analysis.fit(**instance.__dict__)

                # TODO: Use multinest to provide best model

                if likelihood > self.max_likelihood:
                    self.max_likelihood = likelihood
                    self.result = Result(instance, likelihood)

                return likelihood

        # noinspection PyUnusedLocal
        def prior(cube, ndim, nparams):

            phys_cube = self.variable.physical_vector_from_hypercube_vector(hypercube_vector=cube)

            for i in range(self.variable.total_parameters):
                cube[i] = phys_cube[i]

            return cube

        fitness_function = Fitness(self.variable.instance_from_physical_vector, self.constant)

        logger.info("Running MultiNest...")
        self.run(fitness_function.__call__, prior, self.variable.total_parameters,
                 outputfiles_basename="{}/".format(self.path), n_live_points=self.n_live_points,
                 const_efficiency_mode=self.const_efficiency_mode,
                 importance_nested_sampling=self.importance_nested_sampling,
                 evidence_tolerance=self.evidence_tolerance, sampling_efficiency=self.sampling_efficiency,
                 null_log_evidence=self.null_log_evidence, n_iter_before_update=self.n_iter_before_update,
                 multimodal=self.multimodal, max_modes=self.max_modes, mode_tolerance=self.mode_tolerance,
                 seed=self.seed,
                 verbose=self.verbose, resume=self.resume, context=self.context, write_output=self.write_output,
                 log_zero=self.log_zero, max_iter=self.max_iter, init_MPI=self.init_MPI)
        logger.info("MultiNest complete")

        constant = self.most_likely_instance_from_summary()
        for key, value in fitness_function.constant.__dict__.items():
            setattr(constant, key, value)
        likelihood = self.max_likelihood_from_summary()
        variable = self.variable.mapper_from_gaussian_tuples(self.gaussian_priors_at_sigma_limit(self.sigma_limit))

        result = Result(constant=constant, likelihood=likelihood, variable=variable)

        return result

    def open_summary_file(self):

        summary = open(self.file_summary)

        expected_parameters = (len(summary.readline()) - 113) / 112

        if expected_parameters != self.variable.total_parameters:
            raise exc.MultiNestException(
                'The file_summary file has a different number of parameters than the input model')

        return summary

    def read_vector_from_summary(self, number_entries, offset):

        summary = self.open_summary_file()

        summary.seek(0)
        summary.read(2 + offset * self.variable.total_parameters)
        vector = []
        for param in range(number_entries):
            vector.append(float(summary.read(28)))

        summary.close()

        return vector

    def most_probable_from_summary(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt' file which nlo from a \
        multinest analysis.

        This file stores the parameters of the most probable model in the first half of entries and the most likely
        model in the second half of entries. The offset parameter is used to start at the desired model.

        """
        return self.read_vector_from_summary(number_entries=self.variable.total_parameters, offset=0)

    def most_likely_from_summary(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt' file which nlo from a \
        multinest analysis.

        This file stores the parameters of the most probable model in the first half of entries and the most likely
        model in the second half of entries. The offset parameter is used to start at the desired model.
        """
        return self.read_vector_from_summary(number_entries=self.variable.total_parameters, offset=56)

    def max_likelihood_from_summary(self):
        return self.read_vector_from_summary(number_entries=2, offset=112)[0]

    def max_log_likelihood_from_summary(self):
        return self.read_vector_from_summary(number_entries=2, offset=112)[1]

    def most_probable_instance_from_summary(self):
        most_probable = self.most_probable_from_summary()
        return self.variable.instance_from_physical_vector(most_probable)

    def most_likely_instance_from_summary(self):
        most_likely = self.most_likely_from_summary()
        return self.variable.instance_from_physical_vector(most_likely)

    def gaussian_priors_at_sigma_limit(self, sigma_limit):
        """Compute the Gaussian Priors these results should be initialzed with in the next phase, by taking their \
        most probable values (e.g the means of their PDF) and computing the error at an input sigma_limit.

        Parameters
        -----------
        sigma_limit : float
            The sigma limit within which the PDF is used to estimate errors (e.g. sigma_limit = 1.0 uses 0.6826 of the \
            PDF).
        """

        means = self.most_probable_from_summary()
        uppers = self.model_at_upper_sigma_limit(sigma_limit)
        lowers = self.model_at_lower_sigma_limit(sigma_limit)

        # noinspection PyArgumentList
        sigmas = list(map(lambda mean, upper, lower: max([upper - mean, mean - lower]), means, uppers, lowers))

        return list(map(lambda mean, sigma: (mean, sigma), means, sigmas))

    def model_at_sigma_limit(self, sigma_limit):
        limit = math.erf(0.5 * sigma_limit * math.sqrt(2))
        densities_1d = list(map(lambda p: self.pdf.get1DDensity(p), self.pdf.getParamNames().names))
        return list(map(lambda p: p.getLimits(limit), densities_1d))

    def model_at_upper_sigma_limit(self, sigma_limit):
        """Setup 1D vectors of the upper and lower limits of the multinest nlo.

        These are generated at an input limfrac, which gives the percentage of 1d posterior weighted samples within \
        each parameter estimate

        Parameters
        -----------
        sigma_limit : float
            The sigma limit within which the PDF is used to estimate errors (e.g. sigma_limit = 1.0 uses 0.6826 of the \
            PDF).
        """
        return list(map(lambda param: param[1], self.model_at_sigma_limit(sigma_limit)))

    def model_at_lower_sigma_limit(self, sigma_limit):
        """Setup 1D vectors of the upper and lower limits of the multinest nlo.

        These are generated at an input limfrac, which gives the percentage of 1d posterior weighted samples within \
        each parameter estimate

        Parameters
        -----------
        sigma_limit : float
            The sigma limit within which the PDF is used to estimate errors (e.g. sigma_limit = 1.0 uses 0.6826 of the \
            PDF).
        """
        return list(map(lambda param: param[0], self.model_at_sigma_limit(sigma_limit)))

    def weighted_sample_instance_from_weighted_samples(self, index):
        """Setup a model instance of a weighted sample, including its weight and likelihood.

        Parameters
        -----------
        index : int
            The index of the weighted sample to return.
        """
        model, weight, likelihood = self.weighted_sample_model_from_weighted_samples(index)

        self._weighted_sample_model = model

        return self.variable.instance_from_physical_vector(model), weight, likelihood

    def weighted_sample_model_from_weighted_samples(self, index):
        """From a weighted sample return the model, weight and likelihood hood.

        NOTE: GetDist reads the log likelihood from the weighted_sample.txt file (column 2), which are defined as \
        -2.0*likelihood. This routine converts these back to likelihood.

        Parameters
        -----------
        index : int
            The index of the weighted sample to return.
        """
        return list(self.pdf.samples[index]), self.pdf.weights[index], -0.5 * self.pdf.loglikes[index]



from profiling import profiling_data
from profiling import tools
from analysis import ray_tracing
from analysis import galaxy
from profiles import mass_profiles
from pixelization import pixelization
from autolens import exc
import numpy as np
import pytest
import numba

class Reconstructor(object):

    def __init__(self, mapping, regularization, image_to_pix, sub_to_pix):
        """The matrices and mappings used to linearly invert and fit a data-set.

        Parameters
        -----------
        mapping : ndarray
            The matrix representing the mapping between reconstruction-pixels and weighted_data-pixels.
        regularization : ndarray
            The matrix defining how the reconstruction's pixels are regularized with one another when fitting the
            weighted_data.
        image_to_pix : ndarray
            The mapping between each image-grid pixel and pixelization-grid pixel.
        sub_to_pix : ndarray
            The mapping between each sub-grid pixel and pixelization-grid sub-pixel.
        """
        self.mapping = mapping
        self.mapping_shape = mapping.shape
        self.regularization = regularization
        self.image_to_pix = image_to_pix
        self.sub_to_pix = sub_to_pix

sub_grid_size = 4
psf_size = (14, 14)

sie = mass_profiles.EllipticalIsothermal(centre=(0.010, 0.032), einstein_radius=1.47, axis_ratio=0.849, phi=73.6)
shear = mass_profiles.ExternalShear(magnitude=0.0663, phi=160.5)
lens_galaxy = galaxy.Galaxy(mass_profile_0=sie, mass_profile_1=shear)

source_pix = galaxy.Galaxy(pixelization=pixelization.RectangularRegConst(shape=(33, 33)))

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size, psf_shape=psf_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size, psf_shape=psf_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size, psf_shape=psf_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size, psf_shape=psf_size)
# ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size, psf_shape=psf_size)

lsst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=lsst.grids)
euclid_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=euclid.grids)
hst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=hst.grids)
hst_up_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=hst_up.grids)
# ao_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=ao.grids)

lsst_recon = lsst_tracer.reconstructors_from_source_plane(lsst.borders, cluster_mask=None)
euclid_recon = euclid_tracer.reconstructors_from_source_plane(euclid.borders, cluster_mask=None)
hst_recon = hst_tracer.reconstructors_from_source_plane(hst.borders, cluster_mask=None)
hst_up_recon = hst_up_tracer.reconstructors_from_source_plane(hst_up.borders, cluster_mask=None)
# ao_recon = ao_tracer.reconstructors_from_source_plane(ao.borders, cluster_mask=None)

lsst_recon = Reconstructor(lsst_recon.mapping, lsst_recon.regularization, lsst_recon.image_to_pix,
                           lsst_recon.sub_to_pix)
euclid_recon = Reconstructor(euclid_recon.mapping, euclid_recon.regularization, euclid_recon.image_to_pix,
                           euclid_recon.sub_to_pix)
hst_recon = Reconstructor(hst_recon.mapping, hst_recon.regularization, hst_recon.image_to_pix,
                           hst_recon.sub_to_pix)
hst_up_recon = Reconstructor(hst_up_recon.mapping, hst_up_recon.regularization, hst_up_recon.image_to_pix,
                           hst_up_recon.sub_to_pix)
# ao_recon = Reconstructor(ao_recon.mapping, ao_recon.regularization, ao_recon.image_to_pix,
 #                           ao_recon.sub_to_pix)

# Function is already jitted

lsst.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(lsst_recon.mapping)
euclid.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(euclid_recon.mapping)
hst.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(hst_recon.mapping)
hst_up.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(hst_up_recon.mapping)
# ao.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(ao_recon.mapping)

@tools.tick_toc_x1
def lsst_solution():
    lsst.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(lsst_recon.mapping)

@tools.tick_toc_x1
def euclid_solution():
    euclid.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(euclid_recon.mapping)

@tools.tick_toc_x1
def hst_solution():
    hst.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(hst_recon.mapping)

@tools.tick_toc_x1
def hst_up_solution():
    hst_up.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(hst_up_recon.mapping)

@tools.tick_toc_x1
def ao_solution():
    ao.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(ao_recon.mapping)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()


import numpy as np
from astropy.io import fits
import os

def numpy_array_to_fits(array, file_path):

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(array, new_hdr)
    hdu.writeto(file_path + '.fits')

def numpy_array_from_fits(file_path, hdu):
    hdu_list = fits.open(file_path + '.fits')
    return np.array(hdu_list[hdu].data)

def compute_residuals(observed_image, model_image):
    """ Calculate the residuals between an observed image and a model of that image, \
    by subtracting the model from the weighted_data.

    Residuals = (Data - Model).

    Parameters
    ----------
    observed_image : ndarray
        Two dimensional array containing observed image weighted_data
    model_image : nddarray
        Two dimensional array containing model_mapper of the observed image
    """
    return np.subtract(observed_image, model_image)

def compute_variances_from_noise(noise):
    """The variances are the signal_to_noise_ratio (standard deviations) squared."""
    return np.square(noise)

def compute_chi_sq_image(observed_image, model_image, noise):
    """ Calculate the chi-squared image of an observed image, model image and signal_to_noise_ratio map.

    The signal_to_noise_ratio map gives the standard deviation of each image pixel, and is converted to variances (by squaring) in the
     chi squared calculation.

     Chi_Sqs = ((Data - Model)/Noise)**2.0 = (Residuals/Noise)**2.0 = (Residuals**2.0/Variances).

    Parameters
    ----------
    observed_image : ndarray
        Two dimensional array containing observed image weighted_data
    model_image : nddarray
        Two dimensional array containing model_mapper of the observed image
    noise : nddarray
        Two dimensional array containing the signal_to_noise_ratio (standard deviation) in each observed image pixel.
    """
    residuals = compute_residuals(observed_image, model_image)
    variances = compute_variances_from_noise(noise)
    return np.divide(np.square(residuals), variances)

def compute_likelihood(observed_image, model_image, noise):
    """ Calculate the likelihood image of an observed image, model image and signal_to_noise_ratio map. The likelihood is defined as
    the sum of chi squared values multiplied by 0.5.

    The signal_to_noise_ratio map gives the standard deviation of each image pixel, and is converted to variances (by squaring) in the
    chi squared calculation.

    NOTE1 : For efficiency, this routine uses only numpy routines in one big calculation, thus it is 'unpythonic'.

    Parameters
    ----------
    observed_data : ndarray
        Two dimensional array containing the observed image weighted_data
    model_image : nddarray
        Two dimensional array containing the model of the observed image
    noise : nddarray
        Two dimensional array containing the signal_to_noise_ratio (standard deviation) in each observed image pixel
    """
    return -0.5 * np.sum(np.square(np.divide(np.subtract(observed_image, model_image), noise)))
